from typing import Self, cast, Sequence, Final
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    NumPyState,
    NumPyStateSequence,
    NumPyStateBatch,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    NumPyUnicycleObstacleStatesHistory,
    UnicycleState,
    UnicycleStateSequence,
    UnicycleStateBatch,
    UnicycleControlInputSequence,
    UnicycleControlInputBatch,
    UnicyclePositions,
    UnicycleD_x,
    UNICYCLE_D_X,
    UnicycleD_u,
    UNICYCLE_D_U,
    UnicycleD_o,
    DynamicalModel,
    ObstacleModel,
    ObstacleStateEstimator,
    EstimatedObstacleStates,
    NumPyGaussianBelief,
    NumPyNoiseCovarianceDescription,
    NumPyNoiseModelProvider,
)
from faran.filters import (
    NumPyExtendedKalmanFilter,
    NumPyUnscentedKalmanFilter,
    numpy_kalman_filter,
)
from faran.obstacles import NumPyObstacle2dPoses
from faran.models.common import SMALL_UNCERTAINTY, LARGE_UNCERTAINTY
from faran.models.basic import invalid_obstacle_filter_from

from numtypes import D, array

from jaxtyping import Float

import numpy as np

NO_LIMITS: Final = (float("-inf"), float("inf"))

UNICYCLE_ESTIMATION_D_X: Final = 5
UNICYCLE_OBSERVATION_D_O: Final = 3

type UnicycleEstimationD_x = D[5]
type UnicycleObservationD_o = D[3]

type StateArray = Float[Array, " UnicycleD_x"]
type ControlInputSequenceArray = Float[Array, "T UnicycleD_u"]
type StateBatchArray = Float[Array, "T UnicycleD_x M"]
type ControlInputBatchArray = Float[Array, "T UnicycleD_u M"]

type StatesAtTimeStep = Float[Array, "UnicycleD_x M"]
type ControlInputsAtTimeStep = Float[Array, "UnicycleD_u M"]

type EstimationStateCovarianceArray = Float[
    Array, "UnicycleEstimationD_x UnicycleEstimationD_x"
]
type ProcessNoiseCovarianceArray = Float[
    Array, "UnicycleEstimationD_x UnicycleEstimationD_x"
]
type ObservationNoiseCovarianceArray = Float[
    Array, "UnicycleObservationD_o UnicycleObservationD_o"
]
type ObservationMatrix = Float[Array, "UnicycleObservationD_o UnicycleEstimationD_x"]

type UnicycleGaussianBelief = NumPyGaussianBelief
type NumPyUnicycleObstacleCovariances = Float[
    Array, "UnicycleEstimationD_x UnicycleEstimationD_x K"
]
type KalmanFilter = NumPyExtendedKalmanFilter | NumPyUnscentedKalmanFilter


@jaxtyped
@dataclass(frozen=True)
class NumPyUnicycleState(UnicycleState, NumPyState):
    """Kinematic unicycle state: [x, y, heading]."""

    _array: StateArray

    @staticmethod
    def create(*, x: float, y: float, heading: float) -> "NumPyUnicycleState":
        """Creates a NumPy unicycle state from individual state components."""
        return NumPyUnicycleState(array([x, y, heading], shape=(UNICYCLE_D_X,)))

    def __array__(self, dtype: DataType | None = None) -> StateArray:
        return self.array

    @property
    def dimension(self) -> UnicycleD_x:
        return cast(UnicycleD_x, self.array.shape[0])

    @property
    def x(self) -> float:
        return self.array[0]

    @property
    def y(self) -> float:
        return self.array[1]

    @property
    def heading(self) -> float:
        return self.array[2]

    @property
    def array(self) -> StateArray:
        return self._array


@dataclass(kw_only=True, frozen=True)
class NumPyUnicycleStateSequence(UnicycleStateSequence, NumPyStateSequence):
    batch: "NumPyUnicycleStateBatch"
    rollout: int

    @staticmethod
    def of_states(states: Sequence[NumPyUnicycleState]) -> "NumPyUnicycleStateSequence":
        assert len(states) > 0, "States sequence must not be empty."

        array = np.stack([state.array for state in states], axis=0)[:, :, np.newaxis]

        return NumPyUnicycleStateSequence(
            batch=NumPyUnicycleStateBatch.wrap(array), rollout=0
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T UnicycleD_x"]:
        return self.array

    def step(self, index: int) -> NumPyUnicycleState:
        return NumPyUnicycleState(self.array[index])

    def batched(self) -> "NumPyUnicycleStateBatch":
        return NumPyUnicycleStateBatch.wrap(self.array[..., np.newaxis])

    def x(self) -> Float[Array, " T"]:
        return self.array[:, 0]

    def y(self) -> Float[Array, " T"]:
        return self.array[:, 1]

    def heading(self) -> Float[Array, " T"]:
        return self.array[:, 2]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_x:
        return cast(UnicycleD_x, self.array.shape[1])

    @property
    def array(self) -> Float[Array, "T UnicycleD_x"]:
        return self.batch.array[:, :, self.rollout]


@jaxtyped
@dataclass(frozen=True)
class NumPyUnicycleStateBatch(UnicycleStateBatch, NumPyStateBatch):
    _array: StateBatchArray

    @staticmethod
    def wrap(
        array: StateBatchArray,
    ) -> "NumPyUnicycleStateBatch":
        return NumPyUnicycleStateBatch(array)

    @staticmethod
    def of_states(states: Sequence[NumPyUnicycleState]) -> "NumPyUnicycleStateBatch":
        assert len(states) > 0, "States sequence must not be empty."

        array = np.stack([state.array for state in states], axis=0)[:, :, np.newaxis]

        return NumPyUnicycleStateBatch(array)

    def __array__(self, dtype: DataType | None = None) -> StateBatchArray:
        return self.array

    def heading(self) -> Float[Array, "T M"]:
        return self.array[:, 2, :]

    def rollout(self, index: int) -> NumPyUnicycleStateSequence:
        return NumPyUnicycleStateSequence(batch=self, rollout=index)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_x:
        return cast(UnicycleD_x, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def positions(self) -> "NumPyUnicyclePositions":
        return NumPyUnicyclePositions(batch=self)

    @property
    def array(self) -> StateBatchArray:
        return self._array


@dataclass(frozen=True)
class NumPyUnicyclePositions(UnicyclePositions):
    batch: NumPyUnicycleStateBatch

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 M"]:
        return self.batch.array[:, :2, :]

    def x(self) -> Float[Array, "T M"]:
        return self.batch.array[:, 0, :]

    def y(self) -> Float[Array, "T M"]:
        return self.batch.array[:, 1, :]

    @property
    def array(self) -> Float[Array, "T 2 M"]:
        return self.batch.array[:, :2, :]

    @property
    def horizon(self) -> int:
        return self.batch.horizon

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def rollout_count(self) -> int:
        return self.batch.rollout_count


@jaxtyped
@dataclass(frozen=True)
class NumPyUnicycleControlInputSequence(
    UnicycleControlInputSequence, NumPyControlInputSequence
):
    """Control inputs: [linear velocity, angular velocity]."""

    _array: ControlInputSequenceArray

    @staticmethod
    def zeroes(horizon: int) -> "NumPyUnicycleControlInputSequence":
        """Creates a zeroed control input sequence for the given horizon."""
        return NumPyUnicycleControlInputSequence(np.zeros((horizon, UNICYCLE_D_U)))

    def __array__(self, dtype: DataType | None = None) -> ControlInputSequenceArray:
        return self.array

    def similar(
        self, *, array: Float[Array, "L UnicycleD_u"]
    ) -> "Self | NumPyUnicycleControlInputSequence":
        return self.__class__(array)

    def linear_velocities(self) -> Float[Array, " T"]:
        return self.array[:, 0]

    def angular_velocities(self) -> Float[Array, " T"]:
        return self.array[:, 1]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_u:
        return cast(UnicycleD_u, self.array.shape[1])

    @property
    def array(self) -> ControlInputSequenceArray:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyUnicycleControlInputBatch(UnicycleControlInputBatch, NumPyControlInputBatch):
    _array: ControlInputBatchArray

    @staticmethod
    def zero(
        *, horizon: int, rollout_count: int = 1
    ) -> "NumPyUnicycleControlInputBatch":
        return NumPyUnicycleControlInputBatch(
            np.zeros((horizon, UNICYCLE_D_U, rollout_count))
        )

    @staticmethod
    def create(
        *, array: Float[Array, "T UnicycleD_u M"]
    ) -> "NumPyUnicycleControlInputBatch":
        return NumPyUnicycleControlInputBatch(array)

    @staticmethod
    def of(
        sequence: NumPyUnicycleControlInputSequence,
    ) -> "NumPyUnicycleControlInputBatch":
        return NumPyUnicycleControlInputBatch(sequence.array[..., np.newaxis])

    def __array__(self, dtype: DataType | None = None) -> ControlInputBatchArray:
        return self.array

    def linear_velocity(self) -> Float[Array, "T M"]:
        return self.array[:, 0, :]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleD_u:
        return cast(UnicycleD_u, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> ControlInputBatchArray:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyUnicycleObstacleStates:
    _array: Float[Array, "UnicycleD_o K"]

    @staticmethod
    def wrap(array: Float[Array, "UnicycleD_o K"]) -> "NumPyUnicycleObstacleStates":
        return NumPyUnicycleObstacleStates(array)

    @staticmethod
    def create(
        *, x: Float[Array, " K"], y: Float[Array, " K"], heading: Float[Array, " K"]
    ) -> "NumPyUnicycleObstacleStates":
        return NumPyUnicycleObstacleStates(np.stack([x, y, heading], axis=0))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "UnicycleD_o K"]:
        return self.array

    def x(self) -> Float[Array, " K"]:
        return self.array[0, :]

    def y(self) -> Float[Array, " K"]:
        return self.array[1, :]

    def heading(self) -> Float[Array, " K"]:
        return self.array[2, :]

    @property
    def dimension(self) -> UnicycleD_o:
        return cast(UnicycleD_o, self.array.shape[0])

    @property
    def count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[Array, "UnicycleD_o K"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyUnicycleObstacleStateSequences:
    _array: Float[Array, "T UnicycleEstimationD_x K"]
    _covariance: Float[Array, "T UnicycleEstimationD_x UnicycleEstimationD_x K"]

    @staticmethod
    def create(
        predictions: Sequence[UnicycleGaussianBelief],
    ) -> "NumPyUnicycleObstacleStateSequences":
        assert len(predictions) > 0, "Predictions sequence must not be empty."

        return NumPyUnicycleObstacleStateSequences(
            _array=np.stack([belief.mean for belief in predictions], axis=0),
            _covariance=np.stack([belief.covariance for belief in predictions], axis=0),
        )

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T UnicycleEstimationD_x K"]:
        return self.array

    def x(self) -> Float[Array, "T K"]:
        return self.array[:, 0, :]

    def y(self) -> Float[Array, "T K"]:
        return self.array[:, 1, :]

    def heading(self) -> Float[Array, "T K"]:
        return self.array[:, 2, :]

    def covariance(
        self,
    ) -> Float[Array, "T UnicycleEstimationD_x UnicycleEstimationD_x K"]:
        return self._covariance

    def pose_covariance(self) -> Float[Array, "T PoseD_o PoseD_o K"]:
        return self._covariance[:, :3, :3, :]

    def pose(self) -> NumPyObstacle2dPoses:
        return NumPyObstacle2dPoses.create(
            x=self.x(),
            y=self.y(),
            heading=self.heading(),
            covariance=self.pose_covariance(),
        )

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> UnicycleEstimationD_x:
        return cast(UnicycleEstimationD_x, self.array.shape[1])

    @property
    def count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[Array, "T UnicycleEstimationD_x K"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyUnicycleObstacleInputs:
    _linear_velocities: Float[Array, " K"]
    _angular_velocities: Float[Array, " K"]

    @staticmethod
    def wrap(
        array: Float[Array, "UnicycleD_u K"],
    ) -> "NumPyUnicycleObstacleInputs":
        linear, angular = array
        return NumPyUnicycleObstacleInputs(
            _linear_velocities=linear, _angular_velocities=angular
        )

    @staticmethod
    def create(
        *, linear_velocities: Float[Array, " K"], angular_velocities: Float[Array, " K"]
    ) -> "NumPyUnicycleObstacleInputs":
        return NumPyUnicycleObstacleInputs(
            _linear_velocities=linear_velocities, _angular_velocities=angular_velocities
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "UnicycleD_u K"]:
        return self._array

    def linear_velocities(self) -> Float[Array, " K"]:
        return self._linear_velocities

    def angular_velocities(self) -> Float[Array, " K"]:
        return self._angular_velocities

    def zeroed(
        self, *, linear_velocity: bool = False, angular_velocity: bool = False
    ) -> "NumPyUnicycleObstacleInputs":
        """Returns a version of the inputs with the specified components zeroed out."""
        return NumPyUnicycleObstacleInputs(
            _linear_velocities=np.zeros_like(self._linear_velocities)
            if linear_velocity
            else self._linear_velocities,
            _angular_velocities=np.zeros_like(self._angular_velocities)
            if angular_velocity
            else self._angular_velocities,
        )

    @property
    def dimension(self) -> UnicycleD_u:
        return UNICYCLE_D_U

    @property
    def count(self) -> int:
        return self._linear_velocities.shape[0]

    @property
    def array(self) -> Float[Array, "UnicycleD_u K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "UnicycleD_u K"]:
        return np.stack([self._linear_velocities, self._angular_velocities], axis=0)


@dataclass(kw_only=True, frozen=True)
class NumPyUnicycleModel(
    DynamicalModel[
        NumPyUnicycleState,
        NumPyUnicycleStateSequence,
        NumPyUnicycleStateBatch,
        NumPyUnicycleControlInputSequence,
        NumPyUnicycleControlInputBatch,
    ],
):
    """Kinematic unicycle with direct velocity control, Euler-integrated with configurable limits."""

    _time_step_size: float
    speed_limits: tuple[float, float]
    angular_velocity_limits: tuple[float, float]

    @staticmethod
    def create(
        *,
        time_step_size: float,
        speed_limits: tuple[float, float] | None = None,
        angular_velocity_limits: tuple[float, float] | None = None,
    ) -> "NumPyUnicycleModel":

        return NumPyUnicycleModel(
            _time_step_size=time_step_size,
            speed_limits=speed_limits if speed_limits is not None else NO_LIMITS,
            angular_velocity_limits=angular_velocity_limits
            if angular_velocity_limits is not None
            else NO_LIMITS,
        )

    def simulate(
        self, inputs: NumPyUnicycleControlInputBatch, initial_state: NumPyUnicycleState
    ) -> NumPyUnicycleStateBatch:
        rollout_count = inputs.rollout_count

        initial = np.stack(
            [
                np.full(rollout_count, initial_state.x),
                np.full(rollout_count, initial_state.y),
                np.full(rollout_count, initial_state.heading),
            ]
        )

        return NumPyUnicycleStateBatch(
            simulate(
                inputs.array,
                initial,
                time_step_size=self.time_step_size,
                speed_limits=self.speed_limits,
                angular_velocity_limits=self.angular_velocity_limits,
            )
        )

    def step(
        self, inputs: NumPyUnicycleControlInputSequence, state: NumPyUnicycleState
    ) -> NumPyUnicycleState:
        state_as_rollouts = state.array.reshape(-1, 1)
        first_input = inputs.array[0].reshape(-1, 1)

        return NumPyUnicycleState(
            step(
                state_as_rollouts,
                first_input,
                time_step_size=self.time_step_size,
                speed_limits=self.speed_limits,
                angular_velocity_limits=self.angular_velocity_limits,
            )[:, 0]
        )

    def forward(
        self, inputs: NumPyUnicycleControlInputSequence, state: NumPyUnicycleState
    ) -> NumPyUnicycleStateSequence:
        return self.simulate(NumPyUnicycleControlInputBatch.of(inputs), state).rollout(
            0
        )

    @property
    def time_step_size(self) -> float:
        return self._time_step_size


@dataclass(frozen=True)
class NumPyUnicycleStateEstimationModel:
    """Kinematic unicycle model used for obstacle state estimation."""

    time_step_size: float
    initial_state_covariance: EstimationStateCovarianceArray

    @staticmethod
    def create(
        *,
        time_step_size: float,
        initial_state_covariance: EstimationStateCovarianceArray | None = None,
    ) -> "NumPyUnicycleStateEstimationModel":
        if initial_state_covariance is None:
            initial_state_covariance = (
                NumPyUnicycleStateEstimationModel.default_initial_state_covariance()
            )

        return NumPyUnicycleStateEstimationModel(
            time_step_size=time_step_size,
            initial_state_covariance=initial_state_covariance,
        )

    @staticmethod
    def default_initial_state_covariance() -> EstimationStateCovarianceArray:
        # NOTE: Sure of pose, unsure of velocities.
        return cast(
            EstimationStateCovarianceArray,
            np.diag(
                [
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                ]
            ),
        )

    def __call__(self, state: Float[Array, "D_x K"]) -> Float[Array, "D_x K"]:
        dt = self.time_step_size
        x, y, theta, v, omega = state

        return np.array(
            [
                x + v * np.cos(theta) * dt,
                y + v * np.sin(theta) * dt,
                theta + omega * dt,
                v,
                omega,
            ]
        )

    def jacobian(self, state: Float[Array, "D_x K"]) -> Float[Array, "D_x D_x K"]:
        D_x, K = state.shape
        dt = self.time_step_size
        x, y, theta, v, omega = state

        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)

        jacobian = np.zeros((D_x, D_x, K))

        # Partial derivatives for x
        jacobian[0, 0, :] = 1  # ∂x_next/∂x
        jacobian[0, 2, :] = -v * sin_theta * dt  # ∂x_next/∂theta
        jacobian[0, 3, :] = cos_theta * dt  # ∂x_next/∂v

        # Partial derivatives for y
        jacobian[1, 1, :] = 1  # ∂y_next/∂y
        jacobian[1, 2, :] = v * cos_theta * dt  # ∂y_next/∂theta
        jacobian[1, 3, :] = sin_theta * dt  # ∂y_next/∂v

        # Partial derivatives for theta
        jacobian[2, 2, :] = 1  # ∂theta_next/∂theta
        jacobian[2, 4, :] = dt  # ∂theta_next/∂omega

        # Partial derivatives for v and omega are identity since they are assumed constant
        jacobian[3, 3, :] = 1  # ∂v_next/∂v
        jacobian[4, 4, :] = 1  # ∂omega_next/∂omega

        return jacobian

    def observations_from(
        self, history: NumPyUnicycleObstacleStatesHistory
    ) -> Float[Array, "T UnicycleObservationD_o K"]:
        return np.stack([history.x(), history.y(), history.heading()], axis=1)

    def states_from(
        self, belief: UnicycleGaussianBelief
    ) -> NumPyUnicycleObstacleStates:
        return NumPyUnicycleObstacleStates.wrap(belief.mean[:3, :])

    def inputs_from(
        self, belief: UnicycleGaussianBelief
    ) -> NumPyUnicycleObstacleInputs:
        return NumPyUnicycleObstacleInputs.wrap(belief.mean[3:, :])

    def initial_belief_from(
        self,
        *,
        states: NumPyUnicycleObstacleStates,
        inputs: NumPyUnicycleObstacleInputs,
        covariances: NumPyUnicycleObstacleCovariances | None = None,
    ) -> UnicycleGaussianBelief:
        augmented = np.concatenate([states.array, inputs.array], axis=0)

        if covariances is None:
            # NOTE: No covariance means we are "certain" about the states.
            covariances = np.broadcast_to(  # type: ignore
                np.eye(UNICYCLE_ESTIMATION_D_X)[:, :, np.newaxis] * SMALL_UNCERTAINTY,
                (UNICYCLE_ESTIMATION_D_X, UNICYCLE_ESTIMATION_D_X, states.count),
            ).copy()

        return NumPyGaussianBelief(mean=augmented, covariance=covariances)

    @cached_property
    def observation_matrix(self) -> ObservationMatrix:
        # NOTE: We observe the pose directly.
        return np.array(
            [
                [1, 0, 0, 0, 0],
                [0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0],
            ]
        )


@dataclass(kw_only=True, frozen=True)
class NumPyUnicycleObstacleModel(
    ObstacleModel[
        NumPyUnicycleObstacleStatesHistory,
        NumPyUnicycleObstacleStates,
        NumPyUnicycleObstacleInputs,
        NumPyUnicycleObstacleCovariances,
        NumPyUnicycleObstacleStateSequences,
    ]
):
    """Propagates unicycle kinematics forward given states and control inputs."""

    model: NumPyUnicycleStateEstimationModel
    process_noise_covariance: ProcessNoiseCovarianceArray
    predictor: KalmanFilter

    @staticmethod
    def unscented(
        *,
        time_step_size: float,
        process_noise_covariance: NumPyNoiseCovarianceDescription = 1e-3,
        sigma_point_spread: float = 1.0,
        prior_knowledge: float = 2.0,
    ) -> "NumPyUnicycleObstacleModel":
        """Creates a unicycle obstacle model for obstacle state prediction. This
        model uses the unscented transform for propagating state information through
        the nonlinear unicycle dynamics.

        Args:
            time_step_size: Time step size for state propagation.
            process_noise_covariance: The process noise covariance, either as a
                full covariance array, a diagonal covariance vector, or a scalar
                variance representing isotropic noise.
            sigma_point_spread: Spread of sigma points (α) for the unscented
                transform, controlling how far the sigma points are from the mean.
            prior_knowledge: Prior knowledge (β) about the state distribution
                for the unscented transform.
        """
        return NumPyUnicycleObstacleModel(
            model=NumPyUnicycleStateEstimationModel.create(
                time_step_size=time_step_size
            ),
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=UNICYCLE_ESTIMATION_D_X
            ),
            predictor=NumPyUnscentedKalmanFilter.create(
                alpha=sigma_point_spread, beta=prior_knowledge
            ),
        )

    def forward(
        self,
        *,
        states: NumPyUnicycleObstacleStates,
        inputs: NumPyUnicycleObstacleInputs,
        covariances: NumPyUnicycleObstacleCovariances | None,
        horizon: int,
    ) -> NumPyUnicycleObstacleStateSequences:
        beliefs = []
        last = self.model.initial_belief_from(
            states=states, inputs=inputs, covariances=covariances
        )

        for _ in range(horizon):
            beliefs.append(
                last := self.predictor.predict(
                    belief=last,  # type: ignore
                    state_transition=self.model,
                    process_noise_covariance=self.process_noise_covariance,
                )
            )

        return NumPyUnicycleObstacleStateSequences.create(beliefs)


@dataclass(frozen=True)
class NumPyFiniteDifferenceUnicycleStateEstimator(
    ObstacleStateEstimator[
        NumPyUnicycleObstacleStatesHistory,
        NumPyUnicycleObstacleStates,
        NumPyUnicycleObstacleInputs,
    ]
):
    time_step_size: float

    @staticmethod
    def create(
        *, time_step_size: float
    ) -> "NumPyFiniteDifferenceUnicycleStateEstimator":
        return NumPyFiniteDifferenceUnicycleStateEstimator(
            time_step_size=time_step_size
        )

    def estimate_from(
        self, history: NumPyUnicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        NumPyUnicycleObstacleStates, NumPyUnicycleObstacleInputs, None
    ]:
        """Estimates current states and inputs from position/heading history using finite differences.

        Computes the following quantities from the unicycle model (requires T ≥ 2, otherwise returns zeros):

        **Linear velocity**:
            Projection of displacement onto the heading direction (negative for reverse):
            $$v_t = \\frac{(x_t - x_{t-1}) \\cos(\\theta_t) + (y_t - y_{t-1}) \\sin(\\theta_t)}{\\Delta t}$$

        **Angular velocity**:
            Change in heading over time:
            $$\\omega_t = \\frac{\\theta_t - \\theta_{t-1}}{\\Delta t}$$

        Args:
            history: History of observed poses with at least one entry.
        """
        assert history.horizon > 0, "History must contain at least one state."

        filter_invalid = invalid_obstacle_filter_from(history, check_recent=2)
        linear_velocities = self.estimate_speeds_from(history)
        angular_velocities = self.estimate_angular_velocities_from(history)

        return EstimatedObstacleStates(
            states=NumPyUnicycleObstacleStates.create(
                x=history.x()[-1],
                y=history.y()[-1],
                heading=history.heading()[-1],
            ),
            inputs=NumPyUnicycleObstacleInputs.create(
                linear_velocities=filter_invalid(linear_velocities),
                angular_velocities=filter_invalid(angular_velocities),
            ),
            covariance=None,
        )

    def estimate_speeds_from(
        self, history: NumPyUnicycleObstacleStatesHistory
    ) -> Float[Array, " K"]:
        if history.horizon < 2:
            return np.zeros((history.count,))

        x = history.x()
        y = history.y()
        heading = history.heading()

        return self._estimate_speeds_from(
            x_current=x[-1],
            y_current=y[-1],
            x_previous=x[-2],
            y_previous=y[-2],
            heading_current=heading[-1],
        )

    def estimate_angular_velocities_from(
        self, history: NumPyUnicycleObstacleStatesHistory
    ) -> Float[Array, " K"]:
        if history.horizon < 2:
            return np.zeros((history.count,))

        heading = history.heading()

        return self._estimate_angular_velocities_from(
            heading_current=heading[-1],
            heading_previous=heading[-2],
        )

    def _estimate_speeds_from(
        self,
        *,
        x_current: Float[Array, " K"],
        y_current: Float[Array, " K"],
        x_previous: Float[Array, " K"],
        y_previous: Float[Array, " K"],
        heading_current: Float[Array, " K"],
    ) -> Float[Array, " K"]:
        delta_x = x_current - x_previous
        delta_y = y_current - y_previous

        speeds = (
            delta_x * np.cos(heading_current) + delta_y * np.sin(heading_current)
        ) / self.time_step_size

        return speeds

    def _estimate_angular_velocities_from(
        self,
        *,
        heading_current: Float[Array, " K"],
        heading_previous: Float[Array, " K"],
    ) -> Float[Array, " K"]:
        angular_velocities = (heading_current - heading_previous) / self.time_step_size

        return angular_velocities


@dataclass(frozen=True)
class NumPyKfUnicycleStateEstimator(
    ObstacleStateEstimator[
        NumPyUnicycleObstacleStatesHistory,
        NumPyUnicycleObstacleStates,
        NumPyUnicycleObstacleInputs,
        NumPyUnicycleObstacleCovariances,
    ]
):
    """Kalman Filter state estimator for unicycle model obstacles."""

    process_noise_covariance: ProcessNoiseCovarianceArray
    observation_noise_covariance: ObservationNoiseCovarianceArray
    model: NumPyUnicycleStateEstimationModel
    estimator: KalmanFilter

    @staticmethod
    def ekf(
        *,
        time_step_size: float,
        process_noise_covariance: NumPyNoiseCovarianceDescription,
        observation_noise_covariance: NumPyNoiseCovarianceDescription,
        initial_state_covariance: EstimationStateCovarianceArray | None = None,
        noise_model: NumPyNoiseModelProvider | None = None,
    ) -> "NumPyKfUnicycleStateEstimator":
        """Creates an EKF state estimator for the unicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for unobserved velocities.
        """
        return NumPyKfUnicycleStateEstimator(
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=UNICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=UNICYCLE_OBSERVATION_D_O
            ),
            model=NumPyUnicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=NumPyExtendedKalmanFilter.create(noise_model=noise_model),
        )

    @staticmethod
    def ukf(
        *,
        time_step_size: float,
        process_noise_covariance: NumPyNoiseCovarianceDescription,
        observation_noise_covariance: NumPyNoiseCovarianceDescription,
        initial_state_covariance: EstimationStateCovarianceArray | None = None,
        noise_model: NumPyNoiseModelProvider | None = None,
    ) -> "NumPyKfUnicycleStateEstimator":
        """Creates a UKF state estimator for the unicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for unobserved velocities.
        """
        return NumPyKfUnicycleStateEstimator(
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=UNICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=UNICYCLE_OBSERVATION_D_O
            ),
            model=NumPyUnicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=NumPyUnscentedKalmanFilter.create(noise_model=noise_model),
        )

    def estimate_from(
        self, history: NumPyUnicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        NumPyUnicycleObstacleStates,
        NumPyUnicycleObstacleInputs,
        NumPyUnicycleObstacleCovariances,
    ]:
        estimate = self.estimator.filter(
            self.model.observations_from(history),
            initial_state_covariance=self.model.initial_state_covariance,
            state_transition=self.model,
            process_noise_covariance=self.process_noise_covariance,
            observation_noise_covariance=self.observation_noise_covariance,
            observation_matrix=self.model.observation_matrix,
        )

        return EstimatedObstacleStates(
            states=self.model.states_from(estimate),
            inputs=self.model.inputs_from(estimate),
            covariance=estimate.covariance,
        )


def simulate(
    inputs: ControlInputBatchArray,
    initial: StatesAtTimeStep,
    *,
    time_step_size: float,
    speed_limits: tuple[float, float],
    angular_velocity_limits: tuple[float, float],
) -> StateBatchArray:
    horizon = inputs.shape[0]
    rollout_count = inputs.shape[2]
    states = np.zeros((horizon, UNICYCLE_D_X, rollout_count))
    current = initial

    for t in range(horizon):
        current = step(
            current,
            inputs[t],
            time_step_size=time_step_size,
            speed_limits=speed_limits,
            angular_velocity_limits=angular_velocity_limits,
        )
        states[t] = current

    return states


def step(
    state: StatesAtTimeStep,
    control: ControlInputsAtTimeStep,
    *,
    time_step_size: float,
    speed_limits: tuple[float, float],
    angular_velocity_limits: tuple[float, float],
) -> StatesAtTimeStep:
    x, y, theta = state[0], state[1], state[2]
    v, omega = control[0], control[1]
    linear_velocity = np.clip(v, *speed_limits)
    angular_velocity = np.clip(omega, *angular_velocity_limits)

    new_x = x + linear_velocity * np.cos(theta) * time_step_size
    new_y = y + linear_velocity * np.sin(theta) * time_step_size
    new_theta = theta + angular_velocity * time_step_size

    return np.stack([new_x, new_y, new_theta])



from typing import Any
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    StateSequence,
    SimulationData,
    Metric,
    Positions,
    LongitudinalPositions,
    PositionExtractor,
    Trajectory,
)
from faran.collectors import access

from numtypes import array
from jaxtyping import Float, Bool

import numpy as np


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class TaskCompletionMetricResult:
    """Results of the task completion metric.

    Attributes:
        completion: Boolean array indicating whether the task was completed at each time step.
        completion_time: Time at which the task was first completed, or inf if not completed.
        stretch: Ratio of traversed distance to optimal distance.
        completed_part: Proportion of the reference trajectory completed, including lap count.
            Progress within the current lap is clamped to [0, 1].
            Laps are counted when longitudinal position wraps around (drops by >50% of path length).
    """

    completion: Bool[Array, " T"]
    completion_time: float
    stretch: float
    completed_part: float

    @cached_property
    def completed(self) -> bool:
        return bool(self.completion.any())


@dataclass(kw_only=True, frozen=True)
class TaskCompletionMetric[
    StateBatchT,
    PositionsT: Positions,
    LongitudinalT: LongitudinalPositions,
](Metric[TaskCompletionMetricResult]):
    """Metric evaluating whether the ego vehicle reached the goal position."""

    reference: Trajectory[Any, Any, PositionsT, Any, LongitudinalT]
    goal_position: Float[Array, "2"]
    optimal_distance: float
    path_length: float
    distance_threshold: float
    lap_detection_threshold: float
    time_step_size: float
    position_extractor: PositionExtractor[StateBatchT, PositionsT]

    @staticmethod
    def create[S, P: Positions, L: LongitudinalPositions](
        *,
        reference: Trajectory[Any, Any, P, Any, L],
        distance_threshold: float,
        time_step_size: float,
        position_extractor: PositionExtractor[S, P],
        lap_detection_threshold: float = 0.5,
    ) -> "TaskCompletionMetric[S, P, L]":
        assert reference.natural_length > 0.0, (
            f"Reference trajectory must have positive length, got {reference.natural_length}."
        )

        return TaskCompletionMetric(
            reference=reference,
            goal_position=array(reference.end, shape=(2,)),
            optimal_distance=reference.natural_length,
            path_length=reference.path_length,
            distance_threshold=distance_threshold,
            time_step_size=time_step_size,
            position_extractor=position_extractor,
            lap_detection_threshold=lap_detection_threshold,
        )

    def compute(self, data: SimulationData) -> TaskCompletionMetricResult:
        states = data(access.states.assume(StateSequence[StateBatchT]).require())
        positions = self.position_extractor(states.batched())
        positions_array = np.asarray(positions)[..., 0]

        return TaskCompletionMetricResult(
            completion=(completion := self._compute_completion(positions_array)),
            completion_time=self._compute_completion_time(completion),
            stretch=self._compute_stretch(positions_array),
            completed_part=self._compute_completed_part(positions),
        )

    @property
    def name(self) -> str:
        return "task-completion"

    def _compute_completion(self, positions: Float[Array, "T 2"]) -> Bool[Array, " T"]:
        distances_from_goal = np.linalg.norm(positions - self.goal_position, axis=1)
        return distances_from_goal <= self.distance_threshold

    def _compute_completion_time(self, completion: Bool[Array, " T"]) -> float:
        return (
            float(np.argmax(completion)) * self.time_step_size
            if completion.any()
            else float("inf")
        )

    def _compute_stretch(self, positions: Float[Array, "_ 2"]) -> float:
        traversed_distance = np.linalg.norm(np.diff(positions, axis=0), axis=1).sum()

        return float(traversed_distance / self.optimal_distance)

    def _compute_completed_part(self, positions: PositionsT) -> float:
        longitudinal = np.asarray(self.reference.longitudinal(positions))[:, 0]
        normalized = longitudinal / self.path_length
        clamped = np.clip(normalized, 0.0, 1.0)

        if len(normalized) < 2:
            return float(clamped[-1])

        differences = np.diff(normalized)
        wrap_arounds = differences < -self.lap_detection_threshold
        lap_count = int(np.sum(wrap_arounds))

        return float(lap_count + clamped[-1])



from typing import Protocol, Self, Any

from faran.types.mppi.common import (
    State,
    StateSequence,
    StateBatch,
    ControlInputSequence,
    ControlInputBatch,
    DynamicalModel,
    Costs,
    CostFunction,
    Sampler,
    UpdateFunction,
    PaddingFunction,
    FilterFunction,
)

from jaxtyping import Array as JaxArray, Float


class JaxState(State, Protocol):
    @property
    def array(self) -> Float[JaxArray, " D_x"]:
        """Returns the underlying JAX array representing the state."""
        ...


class JaxStateSequence[StateBatchT = Any](StateSequence[StateBatchT], Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_x"]:
        """Returns the underlying JAX array representing the state sequence."""
        ...


class JaxStateBatch(StateBatch, Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_x M"]:
        """Returns the underlying JAX array representing the state batch."""
        ...


class JaxControlInputSequence(ControlInputSequence, Protocol):
    def similar(self, *, array: Float[JaxArray, "T D_u"]) -> Self:
        """Creates a new control input sequence similar to this one but with the given
        array as its data.
        """
        ...

    @property
    def array(self) -> Float[JaxArray, "T D_u"]:
        """Returns the underlying JAX array representing the control input sequence."""
        ...


class JaxControlInputBatch(ControlInputBatch, Protocol):
    @property
    def array(self) -> Float[JaxArray, "T D_u M"]:
        """Returns the underlying JAX array representing the control input batch."""
        ...


class JaxCosts(Costs, Protocol):
    def similar(self, *, array: Float[JaxArray, "T M"]) -> Self:
        """Creates new costs similar to this one but with the given array as its data."""
        ...

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        """Returns the underlying JAX array representing the costs."""
        ...


class JaxDynamicalModel[
    StateT,
    StateSequenceT,
    StateBatchT,
    InputSequenceT,
    InputBatchT,
](
    DynamicalModel[StateT, StateSequenceT, StateBatchT, InputSequenceT, InputBatchT],
    Protocol,
): ...


class JaxSampler[InputSequenceT, InputBatchT](
    Sampler[InputSequenceT, InputBatchT], Protocol
): ...


class JaxCostFunction[InputBatchT, StateBatchT, CostsT](
    CostFunction[InputBatchT, StateBatchT, CostsT], Protocol
): ...


class JaxUpdateFunction[InputSequenceT](UpdateFunction[InputSequenceT], Protocol): ...


class JaxPaddingFunction[NominalT, PaddingT](
    PaddingFunction[NominalT, PaddingT], Protocol
): ...


class JaxFilterFunction[InputSequenceT](FilterFunction[InputSequenceT], Protocol): ...



from typing import cast
from dataclasses import dataclass

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    Distance,
    ControlInputBatch,
    CostFunction,
    ObstacleStates,
    ObstacleStateSampler,
    SampleCostFunction,
    NumPyCosts,
    NumPyObstacleStateProvider,
    NumPyObstacleStateSampler,
    NumPyDistanceExtractor,
    NumPyRisk,
    NumPyRiskMetric,
)
from faran.states import NumPySimpleCosts

from numtypes import D
from jaxtyping import Float

import numpy as np


@jaxtyped
@dataclass(frozen=True)
class NumPyDistance(Distance):
    """Pairwise distances between V vehicle parts and N obstacle samples over T time steps and M rollouts."""

    _array: Float[Array, "T V M N"]

    @staticmethod
    def create(array: Float[Array, "T V M N"]) -> "NumPyDistance":
        """Creates a NumPy distance from the given array."""
        return NumPyDistance(array)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T V M N"]:
        return self.array

    @property
    def horizon(self) -> int:
        return self._array.shape[0]

    @property
    def vehicle_parts(self) -> int:
        return self._array.shape[1]

    @property
    def rollout_count(self) -> int:
        return self._array.shape[2]

    @property
    def sample_count(self) -> int:
        return self._array.shape[3]

    @property
    def array(self) -> Float[Array, "T V M N"]:
        return self._array


class NumPyNoMetric:
    """Bypasses risk computation, evaluating collision cost with a single deterministic sample."""

    @staticmethod
    def create() -> "NumPyNoMetric":
        return NumPyNoMetric()

    def compute[StateT, ObstacleStateT, SampledObstacleStateT](
        self,
        cost_function: SampleCostFunction[
            StateT, SampledObstacleStateT, Float[Array, "T M _"]
        ],
        *,
        states: StateT,
        obstacle_states: ObstacleStateT,
        sampler: ObstacleStateSampler[ObstacleStateT, SampledObstacleStateT],
    ) -> NumPyRisk:
        samples = sampler(obstacle_states, count=1)
        return NumPyRisk(cost_function(states=states, samples=samples).squeeze(axis=-1))

    @property
    def name(self) -> str:
        return "No Metric"


@dataclass(kw_only=True, frozen=True)
class NumPyCollisionCost[
    StateT,
    ObstacleStatesT: ObstacleStates,
    SampledObstacleStatesT,
    DistanceT: NumPyDistance,
](CostFunction[ControlInputBatch, StateT, NumPyCosts]):
    """Collision avoidance cost based on distance thresholds to sampled obstacle positions."""

    obstacle_states: NumPyObstacleStateProvider[ObstacleStatesT]
    sampler: NumPyObstacleStateSampler[ObstacleStatesT, SampledObstacleStatesT]
    distance: NumPyDistanceExtractor[StateT, SampledObstacleStatesT, DistanceT]
    distance_threshold: Float[Array, " V"]
    weight: float
    metric: NumPyRiskMetric[StateT, ObstacleStatesT, SampledObstacleStatesT]

    @staticmethod
    def create[S, OS: ObstacleStates, SOS, D: NumPyDistance](
        *,
        obstacle_states: NumPyObstacleStateProvider[OS],
        sampler: NumPyObstacleStateSampler[OS, SOS],
        distance: NumPyDistanceExtractor[S, SOS, D],
        distance_threshold: Float[Array, " V"],
        weight: float,
        metric: NumPyRiskMetric[S, OS, SOS] | None = None,
    ) -> "NumPyCollisionCost[S, OS, SOS, D]":
        return NumPyCollisionCost(
            obstacle_states=obstacle_states,
            sampler=sampler,
            distance=distance,
            distance_threshold=distance_threshold,
            weight=weight,
            metric=metric
            if metric is not None
            else cast(NumPyRiskMetric[S, OS, SOS], NumPyNoMetric()),
        )

    def __call__(self, *, inputs: ControlInputBatch, states: StateT) -> NumPyCosts:
        def cost(
            *, states: StateT, samples: SampledObstacleStatesT
        ) -> Float[Array, "T M _"]:
            cost = (
                self.distance_threshold[np.newaxis, :, np.newaxis, np.newaxis]
                - self.distance(states=states, obstacle_states=samples).array
            )

            return np.clip(cost, 0, None).sum(axis=1)

        horizon, rollouts = inputs.horizon, inputs.rollout_count

        return NumPySimpleCosts(
            (
                np.zeros((horizon, rollouts))
                if (obstacle_states := self.obstacle_states()).count == 0
                else self.weight
                * self.metric.compute(
                    cost,
                    states=states,
                    obstacle_states=obstacle_states,
                    sampler=self.sampler,
                ).array
            )
        )



from typing import Protocol

from faran.types.array import Array
from faran.types.trajectories import NumPyPathParameters, NumPyPositions, NumPyHeadings
from faran.types.costs.common import PositionExtractor

from jaxtyping import Float


class NumPyPathParameterExtractor[StateBatchT](Protocol):
    def __call__(self, states: StateBatchT, /) -> NumPyPathParameters:
        """Extracts path parameters from a batch of states."""
        ...


class NumPyPathVelocityExtractor[InputBatchT](Protocol):
    def __call__(self, inputs: InputBatchT, /) -> Float[Array, "T M"]:
        """Extracts path velocities from a batch of control inputs."""
        ...


class NumPyPositionExtractor[StateBatchT](
    PositionExtractor[StateBatchT, NumPyPositions], Protocol
): ...


class NumPyHeadingExtractor[StateBatchT](Protocol):
    def __call__(self, states: StateBatchT, /) -> NumPyHeadings:
        """Extracts heading angles from a batch of states."""
        ...



from typing import Protocol

from faran.types.array import Array, DataType

from jaxtyping import Float


class IntegratorState(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, " D_x"]:
        """Returns the state as a NumPy array."""
        ...

    @property
    def dimension(self) -> int:
        """Returns the dimension of the state."""
        ...


class IntegratorStateSequence(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x"]:
        """Returns the state sequence as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the state sequence."""
        ...

    @property
    def dimension(self) -> int:
        """State dimension."""
        ...


class IntegratorStateBatch(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x M"]:
        """Returns the states as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the state batch."""
        ...

    @property
    def dimension(self) -> int:
        """State dimension."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts in the batch."""
        ...


class IntegratorControlInputSequence(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u"]:
        """Returns the control input sequence as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the control input sequence."""
        ...

    @property
    def dimension(self) -> int:
        """Control input dimension."""
        ...


class IntegratorControlInputBatch(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u M"]:
        """Returns the control inputs as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the control input batch."""
        ...

    @property
    def dimension(self) -> int:
        """Control input dimension."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts in the batch."""
        ...



from .common import (
    Noise as Noise,
    NoiseModel as NoiseModel,
    NoiseModelProvider as NoiseModelProvider,
)
from .basic import (
    NumPyNoiseCovarianceArrayDescription as NumPyNoiseCovarianceArrayDescription,
    NumPyNoiseCovarianceDescription as NumPyNoiseCovarianceDescription,
    NumPyGaussianBelief as NumPyGaussianBelief,
    NumPyNoiseCovariances as NumPyNoiseCovariances,
    NumPyNoiseModel as NumPyNoiseModel,
    NumPyNoiseModelProvider as NumPyNoiseModelProvider,
)
from .accelerated import (
    JaxNoiseCovarianceArrayDescription as JaxNoiseCovarianceArrayDescription,
    JaxNoiseCovarianceDescription as JaxNoiseCovarianceDescription,
    JaxGaussianBelief as JaxGaussianBelief,
    JaxNoiseCovariances as JaxNoiseCovariances,
    JaxNoiseModel as JaxNoiseModel,
    JaxNoiseModelProvider as JaxNoiseModelProvider,
)



from typing import NamedTuple, overload
from dataclasses import dataclass

from faran.types import (
    DistanceExtractor,
    Array,
    NumPyHeadings,
    NumPyPositions,
    NumPyPositionExtractor,
    NumPyHeadingExtractor,
    NumPySampledObstaclePositions,
    NumPySampledObstacleHeadings,
    NumPySampledObstaclePositionExtractor,
    NumPySampledObstacleHeadingExtractor,
)
from faran.costs.collision import NumPyDistance
from faran.costs.distance.basic import replace_missing
from faran.costs.distance.sat.common import ConvexPolygon, VerticesArray

from numtypes import D
from jaxtyping import Float, Bool

import numpy as np

type V = D[1]
"""Number of ego polygons (fixed to 1 for now)."""


class MaskedObstacleStates(NamedTuple):
    x: Float[Array, "T K N"]
    y: Float[Array, "T K N"]
    heading: Float[Array, "T K N"]
    is_valid: Bool[Array, "T K N"]


@dataclass(frozen=True)
class NumPySatDistanceExtractor[StateT, SampledObstacleStatesT](
    DistanceExtractor[StateT, SampledObstacleStatesT, NumPyDistance]
):
    """
    Computes the signed distances between the ego polygon and obstacle polygons
    using the Separating Axis Theorem (SAT).

    Positive distance means separated, negative means penetrating.
    """

    ego: ConvexPolygon
    obstacle: ConvexPolygon
    positions_from: NumPyPositionExtractor[StateT]
    headings_from: NumPyHeadingExtractor[StateT]
    obstacle_positions_from: NumPySampledObstaclePositionExtractor[
        SampledObstacleStatesT
    ]
    obstacle_headings_from: NumPySampledObstacleHeadingExtractor[SampledObstacleStatesT]

    @staticmethod
    def create[S, SOS](
        *,
        ego: ConvexPolygon,
        obstacle: ConvexPolygon,
        position_extractor: NumPyPositionExtractor[S],
        heading_extractor: NumPyHeadingExtractor[S],
        obstacle_position_extractor: NumPySampledObstaclePositionExtractor[SOS],
        obstacle_heading_extractor: NumPySampledObstacleHeadingExtractor[SOS],
    ) -> "NumPySatDistanceExtractor[S, SOS]":
        return NumPySatDistanceExtractor(
            ego=ego,
            obstacle=obstacle,
            positions_from=position_extractor,
            headings_from=heading_extractor,
            obstacle_positions_from=obstacle_position_extractor,
            obstacle_headings_from=obstacle_heading_extractor,
        )

    def __call__(
        self, *, states: StateT, obstacle_states: SampledObstacleStatesT
    ) -> NumPyDistance:
        obstacle_positions, obstacle_headings = replace_missing(
            positions=self.obstacle_positions_from(obstacle_states),
            headings=self.obstacle_headings_from(obstacle_states),
        )

        return NumPyDistance(
            compute_sat_distances(
                ego_positions=self.positions_from(states),
                ego_headings=self.headings_from(states),
                ego_vertices=self.ego.vertices,
                obstacle_positions=obstacle_positions,
                obstacle_headings=obstacle_headings,
                obstacle_vertices=self.obstacle.vertices,
            )
        )


def compute_sat_distances(
    *,
    ego_positions: NumPyPositions,
    ego_headings: NumPyHeadings,
    ego_vertices: VerticesArray,
    obstacle_positions: NumPySampledObstaclePositions,
    obstacle_headings: NumPySampledObstacleHeadings,
    obstacle_vertices: VerticesArray,
) -> Float[Array, "T V M N"]:
    ego_x = ego_positions.x()
    ego_y = ego_positions.y()
    ego_heading = ego_headings.heading()

    obstacle_x, obstacle_y, obstacle_heading, is_valid = mask_valid_obstacle_states(
        positions=obstacle_positions, headings=obstacle_headings
    )

    T, M = ego_x.shape
    K = obstacle_x.shape[1]
    N = obstacle_x.shape[2]
    V = 1

    if K == 0:
        distances = np.full((T, V, M, N), np.inf)
        return distances

    ego_polygons = transform_polygons_batched(
        ego_vertices,
        x=ego_x,
        y=ego_y,
        heading=ego_heading,
    )

    obstacle_polygons = transform_polygons_batched(
        obstacle_vertices,
        x=obstacle_x,
        y=obstacle_y,
        heading=obstacle_heading,
    )

    distances = sat_distances_batched(
        ego_polygons=ego_polygons,
        obstacle_polygons=obstacle_polygons,
        is_valid=is_valid,
    )

    min_over_obstacles = np.min(distances, axis=2, keepdims=True)
    result = np.moveaxis(min_over_obstacles, 2, 1)

    return result


def mask_valid_obstacle_states(
    *, positions: NumPySampledObstaclePositions, headings: NumPySampledObstacleHeadings
) -> MaskedObstacleStates:
    x = positions.x()
    y = positions.y()
    heading = headings.heading()
    T, K, N = x.shape

    is_valid = ~(np.isnan(x) | np.isnan(y) | np.isinf(x) | np.isinf(y))
    x = np.where(is_valid, x, 0.0)
    y = np.where(is_valid, y, 0.0)
    heading = np.where(is_valid, heading, 0.0)

    return MaskedObstacleStates(x=x, y=y, heading=heading, is_valid=is_valid)


@overload
def transform_polygons_batched(
    local_vertices: VerticesArray,
    *,
    x: Float[Array, "T M"],
    y: Float[Array, "T M"],
    heading: Float[Array, "T M"],
) -> Float[Array, "T M P 2"]: ...


@overload
def transform_polygons_batched(  # pyright: ignore[reportOverlappingOverload]
    local_vertices: VerticesArray,
    *,
    x: Float[Array, "T K N"],
    y: Float[Array, "T K N"],
    heading: Float[Array, "T K N"],
) -> Float[Array, "T K N P 2"]: ...


def transform_polygons_batched(
    local_vertices: VerticesArray,
    *,
    x: Float[Array, "T M"] | Float[Array, "T K N"],
    y: Float[Array, "T M"] | Float[Array, "T K N"],
    heading: Float[Array, "T M"] | Float[Array, "T K N"],
) -> Float[Array, "T M P 2"] | Float[Array, "T K N P 2"]:
    P = local_vertices.shape[0]
    spatial_shape = x.shape

    rotation_matrices = create_rotation_matrices(heading)
    local_expanded = local_vertices.reshape((1,) * len(spatial_shape) + (P, 2))
    rotated = np.einsum("...pj,...ij->...pi", local_expanded, rotation_matrices)

    positions = np.stack([x, y], axis=-1)
    positions_expanded = positions[..., np.newaxis, :]

    return rotated + positions_expanded


@overload
def create_rotation_matrices(
    heading: Float[Array, "T M"],
) -> Float[Array, "T M 2 2"]: ...


@overload
def create_rotation_matrices(  # pyright: ignore[reportOverlappingOverload]
    heading: Float[Array, "T K N"],
) -> Float[Array, "T K N 2 2"]: ...


def create_rotation_matrices(
    heading: Float[Array, "T M"] | Float[Array, "T K N"],
) -> Float[Array, "T M 2 2"] | Float[Array, "T K N 2 2"]:
    cos_h = np.cos(heading)
    sin_h = np.sin(heading)
    return np.stack(
        [
            np.stack([cos_h, -sin_h], axis=-1),
            np.stack([sin_h, cos_h], axis=-1),
        ],
        axis=-2,
    )


def sat_distances_batched(
    *,
    ego_polygons: Float[Array, "T M P_ego 2"],
    obstacle_polygons: Float[Array, "T K N P_obs 2"],
    is_valid: Bool[Array, "T K N"],
) -> Float[Array, "T M K N"]:
    all_axes = gather_separating_axes(ego_polygons, obstacle_polygons)

    ego_expanded = ego_polygons[:, :, np.newaxis, np.newaxis, :, :]
    obs_expanded = obstacle_polygons[:, np.newaxis, :, :, :, :]

    ego_min, ego_max = project_polygons_onto_axes(ego_expanded, all_axes)
    obs_min, obs_max = project_polygons_onto_axes(obs_expanded, all_axes)

    distances = compute_signed_distances_from_projections(
        ego_min, ego_max, obs_min, obs_max
    )

    is_invalid_expanded = ~is_valid[:, np.newaxis, :, :]
    return np.where(is_invalid_expanded, np.inf, distances)


def gather_separating_axes(
    ego_polygons: Float[Array, "T M P_ego 2"],
    obstacle_polygons: Float[Array, "T K N P_obs 2"],
) -> Float[Array, "T M K N L_axes 2"]:
    P_ego = ego_polygons.shape[2]
    P_obs = obstacle_polygons.shape[3]

    ego_normals = get_edge_normals_batched(ego_polygons)
    obs_normals = get_edge_normals_batched(obstacle_polygons)

    ego_normals_expanded = ego_normals[:, :, np.newaxis, np.newaxis, :, :]
    obs_normals_expanded = obs_normals[:, np.newaxis, :, :, :, :]

    T, M = ego_polygons.shape[:2]
    K, N = obstacle_polygons.shape[1:3]

    ego_normals_full = np.broadcast_to(ego_normals_expanded, (T, M, K, N, P_ego, 2))
    obs_normals_full = np.broadcast_to(obs_normals_expanded, (T, M, K, N, P_obs, 2))

    return np.concatenate([ego_normals_full, obs_normals_full], axis=-2)


@overload
def project_polygons_onto_axes(
    polygons: Float[Array, "T M 1 1 P 2"],
    axes: Float[Array, "T M K N L_axes 2"],
) -> tuple[Float[Array, "T M K N L_axes"], Float[Array, "T M K N L_axes"]]: ...


@overload
def project_polygons_onto_axes(  # pyright: ignore[reportOverlappingOverload]
    polygons: Float[Array, "T 1 K N P 2"],
    axes: Float[Array, "T M K N L_axes 2"],
) -> tuple[Float[Array, "T M K N L_axes"], Float[Array, "T M K N L_axes"]]: ...


def project_polygons_onto_axes(
    polygons: Float[Array, "T M 1 1 P 2"] | Float[Array, "T 1 K N P 2"],
    axes: Float[Array, "T M K N L_axes 2"],
) -> tuple[Float[Array, "T M K N L_axes"], Float[Array, "T M K N L_axes"]]:
    projections = np.einsum("...pi,...ai->...ap", polygons, axes)
    return projections.min(axis=-1), projections.max(axis=-1)


def compute_signed_distances_from_projections(
    ego_min: Float[Array, "T M K N L_axes"],
    ego_max: Float[Array, "T M K N L_axes"],
    obs_min: Float[Array, "T M K N L_axes"],
    obs_max: Float[Array, "T M K N L_axes"],
) -> Float[Array, "T M K N"]:
    gap_ab = obs_min - ego_max
    gap_ba = ego_min - obs_max
    gaps = np.maximum(gap_ab, gap_ba)

    overlap = np.minimum(ego_max - obs_min, obs_max - ego_min)
    penetration = -overlap

    is_separated = gaps > 0
    any_separated = np.any(is_separated, axis=-1)

    separation_values = np.where(is_separated, gaps, np.inf)
    min_separation = np.min(separation_values, axis=-1)

    penetration_values = np.where(~is_separated, penetration, -np.inf)
    max_penetration = np.max(penetration_values, axis=-1)

    return np.where(any_separated, min_separation, max_penetration)


@overload
def get_edge_normals_batched(
    vertices: Float[Array, "T M P_ego 2"],
) -> Float[Array, "T M P_ego 2"]: ...


@overload
def get_edge_normals_batched(  # pyright: ignore[reportOverlappingOverload]
    vertices: Float[Array, "T K N P_obs 2"],
) -> Float[Array, "T K N P_obs 2"]: ...


def get_edge_normals_batched(
    vertices: Float[Array, "T M P 2"] | Float[Array, "T K N P 2"],
) -> Float[Array, "T M P 2"] | Float[Array, "T K N P 2"]:
    edges = np.roll(vertices, -1, axis=-2) - vertices
    normals = np.stack([-edges[..., 1], edges[..., 0]], axis=-1)
    norms = np.linalg.norm(normals, axis=-1, keepdims=True)
    return normals / np.where(norms > 1e-10, norms, 1.0)



from typing import Final

from faran.metrics.registry import MetricRegistry
from faran.metrics.collision import CollisionMetric
from faran.metrics.error import MpccErrorMetric
from faran.metrics.task import TaskCompletionMetric
from faran.metrics.constraint import ConstraintViolationMetric
from faran.metrics.comfort import ComfortMetric


class metrics:
    """Factory namespace for creating simulation evaluation metrics."""

    registry: Final = MetricRegistry.of
    collision: Final = CollisionMetric.create
    mpcc_error: Final = MpccErrorMetric.create
    task_completion: Final = TaskCompletionMetric.create
    constraint_violation: Final = ConstraintViolationMetric.create
    comfort: Final = ComfortMetric.create



from typing import cast, Any
from dataclasses import dataclass

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    NumPyState,
    NumPyStateBatch,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    NumPyDynamicalModel,
    NumPyCostFunction,
    NumPySampler,
    NumPyUpdateFunction,
    NumPyPaddingFunction,
    NumPyFilterFunction,
    Mppi,
    DebugData,
    Control,
)
from faran.mppi.common import UseOptimalControlUpdate, NoFilter

from jaxtyping import Float

import numpy as np


@jaxtyped
@dataclass(frozen=True)
class NumPyWeights:
    """Softmax-normalized cost weights used to compute the weighted average over rollouts."""

    _array: Float[Array, " M"]

    def __array__(self, dtype: DataType | None = None) -> Float[Array, " M"]:
        return self._array

    @property
    def rollout_count(self) -> int:
        return self._array.shape[0]

    @property
    def array(self) -> Float[Array, " M"]:
        return self._array


class NumPyZeroPadding(
    NumPyPaddingFunction[NumPyControlInputSequence, NumPyControlInputSequence]
):
    """Fills shifted-out time steps with zero control inputs after horizon advancement."""

    def __call__(
        self, *, nominal_input: NumPyControlInputSequence, padding_size: int
    ) -> NumPyControlInputSequence:
        array = np.zeros((padding_size, nominal_input.dimension))

        return nominal_input.similar(array=array)


@dataclass(kw_only=True, frozen=True)
class NumPyMppi[
    StateT: NumPyState,
    StateBatchT: NumPyStateBatch,
    ControlInputSequenceT: NumPyControlInputSequence,
    ControlInputBatchT: NumPyControlInputBatch,
    ControlInputPaddingT: NumPyControlInputSequence = ControlInputSequenceT,
](Mppi[StateT, ControlInputSequenceT, NumPyWeights]):
    """Sampling-based stochastic optimal controller using the information-theoretic MPPI algorithm."""

    planning_interval: int
    model: NumPyDynamicalModel[
        StateT, Any, StateBatchT, ControlInputSequenceT, ControlInputBatchT
    ]
    cost_function: NumPyCostFunction
    sampler: NumPySampler[ControlInputSequenceT, ControlInputBatchT]
    update_function: NumPyUpdateFunction[ControlInputSequenceT]
    padding_function: NumPyPaddingFunction[ControlInputSequenceT, ControlInputPaddingT]
    filter_function: NumPyFilterFunction[ControlInputSequenceT]

    @staticmethod
    def create[
        S: NumPyState,
        SB: NumPyStateBatch,
        CIS: NumPyControlInputSequence,
        CIB: NumPyControlInputBatch,
        CIP: NumPyControlInputSequence = CIS,
    ](
        *,
        planning_interval: int = 1,
        model: NumPyDynamicalModel[S, Any, SB, CIS, CIB],
        cost_function: NumPyCostFunction,
        sampler: NumPySampler[CIS, CIB],
        update_function: NumPyUpdateFunction[CIS] | None = None,
        padding_function: NumPyPaddingFunction[CIS, CIP] | None = None,
        filter_function: NumPyFilterFunction[CIS] | None = None,
    ) -> "NumPyMppi[S, SB, CIS, CIB, CIP]":
        return NumPyMppi(
            planning_interval=planning_interval,
            model=model,
            cost_function=cost_function,
            sampler=sampler,
            update_function=update_function
            or cast(NumPyUpdateFunction[CIS], UseOptimalControlUpdate()),
            padding_function=padding_function
            or cast(NumPyPaddingFunction[CIS, CIP], NumPyZeroPadding()),
            filter_function=filter_function
            or cast(NumPyFilterFunction[CIS], NoFilter()),
        )

    def __post_init__(self) -> None:
        assert self.planning_interval > 0, "Planning interval must be positive."

    def step(
        self,
        *,
        temperature: float,
        nominal_input: ControlInputSequenceT,
        initial_state: StateT,
    ) -> Control[ControlInputSequenceT, NumPyWeights]:
        """Samples rollouts, evaluates costs, and returns the cost-weighted optimal control."""
        assert temperature > 0.0, "Temperature must be positive."

        samples = self.sampler.sample(around=nominal_input)

        rollouts = self.model.simulate(inputs=samples, initial_state=initial_state)
        costs = self.cost_function(inputs=samples, states=rollouts)
        costs_per_rollout = np.sum(costs.array, axis=0)

        min_cost = np.min(costs_per_rollout)
        exp_costs = np.exp((costs_per_rollout - min_cost) / (-temperature))

        normalizing_constant = exp_costs.sum()
        weights = exp_costs / normalizing_constant

        optimal_control = np.tensordot(samples, weights, axes=([2], [0]))
        optimal_input = self.filter_function(
            optimal_input=nominal_input.similar(array=optimal_control)
        )

        nominal_input = self.update_function(
            nominal_input=nominal_input, optimal_input=optimal_input
        )

        shifted_control = np.concat(
            [
                nominal_input.array[self.planning_interval :],
                self.padding_function(
                    nominal_input=nominal_input, padding_size=self.planning_interval
                ).array,
            ],
            axis=0,
        )

        return Control(
            optimal=optimal_input,
            nominal=nominal_input.similar(array=shifted_control),
            debug=DebugData(trajectory_weights=NumPyWeights(weights)),
        )



from warnings import warn
from typing import Protocol, Sequence, Any, Callable
from dataclasses import dataclass

from faran.types import SimulationData


type OnModifyCallback = Callable[[], None]


class NoCollectedDataWarning(RuntimeWarning):
    """Warning raised when no data has been collected by a registered collector."""


class DataTransformer[InputT, OutputT = Any](Protocol):
    """Protocol for transforming collected data before storage."""

    def __call__(self, data: Sequence[InputT], /) -> OutputT:
        """Transforms the collected data into the desired output format."""
        ...


class Collector[T](Protocol):
    """Protocol for simulation data collectors with modification callbacks."""

    def on_modified(self, callback: OnModifyCallback) -> None:
        """Registers a callback to be invoked when new data is collected."""
        ...

    def collected(self, *, take: int) -> Sequence[T]:
        """Returns the first `take` collected data points."""
        ...

    @property
    def count(self) -> int:
        """Returns the number of collected data points."""
        ...

    @property
    def key(self) -> str:
        """Returns the key of the data being collected."""
        ...

    @property
    def transformer(self) -> DataTransformer[T, Any]:
        """Returns the transformer function to be applied to the collected data."""
        ...


class HasCallbacks(Protocol):
    """Protocol for objects holding a list of modification callbacks."""

    @property
    def _callbacks(self) -> list[OnModifyCallback]:
        """Returns the list of registered modification callbacks."""
        ...


class HasList[T](Protocol):
    """Protocol for objects holding a list of collected items."""

    @property
    def _collected(self) -> list[T]:
        """Returns the list of collected data."""
        ...


class IdentityTransformer[InputT]:
    """Transformer that returns collected data unchanged."""

    def __call__(self, data: Sequence[InputT], /) -> Sequence[InputT]:
        """Returns the input data unchanged."""
        return data


class ModificationNotifierMixin:
    """Mixin that invokes registered callbacks when data is modified."""

    def notify(self: HasCallbacks) -> None:
        for callback in self._callbacks:
            callback()

    def on_modified(self: HasCallbacks, callback: OnModifyCallback) -> None:
        self._callbacks.append(callback)


class ListCollectorMixin:
    """Mixin providing list-backed collected data access."""

    def collected[T](self: HasList[T], *, take: int) -> Sequence[T]:
        return self._collected[:take]

    @property
    def count(self: HasList[Any]) -> int:
        return len(self._collected)


@dataclass
class CollectorRegistry:
    """Registry aggregating multiple collectors and producing unified simulation data."""

    collectors: tuple[Collector, ...]
    _data: SimulationData | None

    @staticmethod
    def of(*args: Collector) -> "CollectorRegistry":
        registry = CollectorRegistry(collectors=(), _data=None)
        registry.collectors = tuple(
            registry._with_callback(collector)
            for collector in args
            if collector is not None
        )

        return registry

    def on_modified(self, callback: OnModifyCallback) -> None:
        for collector in self.collectors:
            collector.on_modified(callback)

    @property
    def data(self) -> SimulationData:
        if self._data is None:
            self._data = self._collect_data()

        return self._data

    def _modified(self) -> None:
        self._data = None

    def _collect_data(self) -> SimulationData:
        for collector in self.collectors:
            if collector.count == 0:
                warn(
                    f"Collector for key '{collector.key}' has not collected any data.",
                    NoCollectedDataWarning,
                )

        collectors = [it for it in self.collectors if it.count > 0]
        time_step_count = min(collector.count for collector in collectors)

        return SimulationData.create(
            {
                collector.key: collector.transformer(
                    collector.collected(take=time_step_count)
                )
                for collector in collectors
            }
        )

    def _with_callback(self, collector: Collector) -> Collector:
        collector.on_modified(self._modified)
        return collector



from .line import (
    NumPyLineTrajectory as NumPyLineTrajectory,
    JaxLineTrajectory as JaxLineTrajectory,
)
from .waypoints import (
    NumPyWaypointsTrajectory as NumPyWaypointsTrajectory,
    JaxWaypointsTrajectory as JaxWaypointsTrajectory,
)
from .factory import trajectory as trajectory



from typing import NamedTuple

from faran.types import (
    Array,
    NumPyNoiseModel,
    NumPyNoiseModelProvider,
    NumPyGaussianBelief,
    NumPyNoiseCovariances,
)

from jaxtyping import Bool, Float

import numpy as np


class NumPyNoiseCovarianceBounds(NamedTuple):
    process: float
    observation: float

    @staticmethod
    def create(*, process: float, observation: float) -> "NumPyNoiseCovarianceBounds":
        """Creates isotropic bounds for noise covariances.

        Args:
            process: Isotropic bound for process noise covariance eigenvalues.
            observation: Isotropic bound for observation noise covariance eigenvalues.
        """
        return NumPyNoiseCovarianceBounds(process=process, observation=observation)


class NumPyClampedNoise[StateT](NamedTuple):
    """Decorator that clamps an inner noise model's eigenvalues to a floor and/or ceiling."""

    inner: NumPyNoiseModel
    floor: NumPyNoiseCovarianceBounds
    ceiling: NumPyNoiseCovarianceBounds

    def __call__(
        self,
        *,
        noise: NumPyNoiseCovariances,
        prediction: NumPyGaussianBelief,
        observation: Float[Array, "D_z K"],
        state: StateT,
    ) -> tuple[NumPyNoiseCovariances, StateT]:
        result, state = self.inner(
            noise=noise, prediction=prediction, observation=observation, state=state
        )
        return NumPyNoiseCovariances(
            process_noise_covariance=clamp_eigenvalues(
                result.process_noise_covariance,
                floor=self.floor.process,
                ceiling=self.ceiling.process,
            ),
            observation_noise_covariance=clamp_eigenvalues(
                result.observation_noise_covariance,
                floor=self.floor.observation,
                ceiling=self.ceiling.observation,
            ),
        ), state

    @property
    def state(self) -> StateT:
        return self.inner.state


class NumPyClampedNoiseProvider[StateT](NamedTuple):
    inner: NumPyNoiseModelProvider[StateT]
    floor: NumPyNoiseCovarianceBounds
    ceiling: NumPyNoiseCovarianceBounds

    @staticmethod
    def decorate[S](
        inner: NumPyNoiseModelProvider[S],
        *,
        floor: NumPyNoiseCovarianceBounds | None = None,
        ceiling: NumPyNoiseCovarianceBounds | None = None,
    ) -> "NumPyClampedNoiseProvider[S]":
        """Creates a noise model provider that clamps the eigenvalues of the
        noise covariances to the specified floor and/or ceiling.

        Args:
            inner: The inner noise model provider to delegate to.
            floor: Isotropic minimum bounds. Eigenvalues of the inner model's
                output will be clamped to be no smaller than these.
            ceiling: Isotropic maximum bounds. Eigenvalues of the inner model's
                output will be clamped to be no larger than these.
        """
        return NumPyClampedNoiseProvider(
            inner=inner,
            floor=floor or NumPyNoiseCovarianceBounds(process=0.0, observation=0.0),
            ceiling=ceiling
            or NumPyNoiseCovarianceBounds(process=np.inf, observation=np.inf),
        )

    def __call__(
        self,
        *,
        obstacle_count: int,
        observation_matrix: Float[Array, "D_z D_x"],
        noise: NumPyNoiseCovariances,
    ) -> NumPyClampedNoise:
        inner_model = self.inner(
            obstacle_count=obstacle_count,
            observation_matrix=observation_matrix,
            noise=noise,
        )
        return NumPyClampedNoise(
            inner=inner_model, floor=self.floor, ceiling=self.ceiling
        )


class NumPyAdaptiveNoiseState(NamedTuple):
    """Circular buffer state for adaptive noise estimation."""

    buffer: Float[Array, "W D_z D_z K"]
    entry_count: Float[Array, " K"]


class NumPyAdaptiveNoise(NamedTuple):
    """Innovation-Based Adaptive Estimation (IAE) for noise covariances."""

    observation_matrix: Float[Array, "D_z D_x"]
    obstacle_count: int
    window_size: int

    def __call__(
        self,
        *,
        noise: NumPyNoiseCovariances,
        prediction: NumPyGaussianBelief,
        observation: Float[Array, "D_z K"],
        state: NumPyAdaptiveNoiseState,
    ) -> tuple[NumPyNoiseCovariances, NumPyAdaptiveNoiseState]:
        if not np.any(valid := valid_obstacle_mask(prediction, observation)):
            return noise, state

        innovation = compute_innovation(
            prediction=prediction,
            observation=observation,
            observation_matrix=self.observation_matrix,
        )
        new_state = compute_updated_state(
            state=state, innovation=innovation, valid=valid
        )

        if not np.any(buffer_full := new_state.entry_count >= self.window_size):
            return noise, new_state

        innovation_matrices = np.median(new_state.buffer, axis=0)
        safe_covariance = np.where(
            np.isnan(prediction.covariance), 0.0, prediction.covariance
        )

        P = safe_covariance.transpose(2, 0, 1)
        V = innovation_matrices.transpose(2, 0, 1)
        H = self.observation_matrix

        kalman_gains = compute_kalman_gain(
            predicted_covariance=P,
            observation_matrix=H,
            observation_noise_covariance=noise.observation_noise_covariance,
        )

        adapted_process = enforce_spd(
            kalman_gains @ V @ kalman_gains.swapaxes(-2, -1)
        ).transpose(1, 2, 0)

        adapted_observation = enforce_spd(V - H @ P @ H.T).transpose(1, 2, 0)

        adapted_noise = aggregate_noise_covariances(
            buffer_full=buffer_full,
            valid=valid,
            adapted_process=adapted_process,
            adapted_observation=adapted_observation,
            noise=noise,
        )

        return adapted_noise, new_state

    @property
    def state(self) -> NumPyAdaptiveNoiseState:
        observation_dimension = self.observation_matrix.shape[0]
        return NumPyAdaptiveNoiseState(
            buffer=np.zeros(
                (
                    self.window_size,
                    observation_dimension,
                    observation_dimension,
                    self.obstacle_count,
                )
            ),
            entry_count=np.zeros(self.obstacle_count, dtype=np.int32),
        )


class NumPyAdaptiveNoiseProvider(NamedTuple):
    window_size: int

    @staticmethod
    def create(*, window_size: int) -> "NumPyAdaptiveNoiseProvider":
        """Creates an innovation-based adaptive estimation model for noise.

        Args:
            window_size: The number of past observations considered, when adapting
                the noise covariances.
        """
        return NumPyAdaptiveNoiseProvider(window_size=window_size)

    def __call__(
        self,
        *,
        obstacle_count: int,
        observation_matrix: Float[Array, "D_z D_x"],
        noise: NumPyNoiseCovariances,
    ) -> NumPyAdaptiveNoise:
        return NumPyAdaptiveNoise(
            observation_matrix=observation_matrix,
            obstacle_count=obstacle_count,
            window_size=self.window_size,
        )


def valid_obstacle_mask(
    prediction: NumPyGaussianBelief, observation: Float[Array, "D_z K"]
) -> Bool[Array, " K"]:
    mean_valid = ~np.any(np.isnan(prediction.mean), axis=0)
    covariance_valid = ~np.any(
        np.isnan(prediction.covariance.reshape(-1, prediction.covariance.shape[2])),
        axis=0,
    )
    observation_valid = ~np.any(np.isnan(observation), axis=0)
    return mean_valid & covariance_valid & observation_valid


def compute_innovation(
    prediction: NumPyGaussianBelief,
    observation: Float[Array, "D_z K"],
    *,
    observation_matrix: Float[Array, "D_z D_x"],
) -> Float[Array, "D_z D_z K"]:
    safe_mean = np.where(np.isnan(prediction.mean), 0.0, prediction.mean)
    safe_observation = np.where(np.isnan(observation), 0.0, observation)
    innovation = safe_observation - observation_matrix @ safe_mean

    return innovation[:, np.newaxis, :] * innovation[np.newaxis, :, :]


def compute_updated_state(
    state: NumPyAdaptiveNoiseState,
    *,
    innovation: Float[Array, "D_z D_z K"],
    valid: Bool[Array, " K"],
) -> NumPyAdaptiveNoiseState:
    window_size = state.buffer.shape[0]
    obstacle_count = valid.shape[0]

    indices = state.entry_count % window_size
    updated_buffer = state.buffer.copy()
    updated_buffer[indices, :, :, np.arange(obstacle_count)] = innovation.transpose(
        2, 0, 1
    )

    return NumPyAdaptiveNoiseState(
        buffer=np.where(
            valid[np.newaxis, np.newaxis, np.newaxis, :],
            updated_buffer,
            state.buffer,
        ),
        entry_count=state.entry_count + valid.astype(np.int32),
    )


def aggregate_noise_covariances(
    *,
    buffer_full: Bool[Array, " K"],
    valid: Bool[Array, " K"],
    adapted_process: Float[Array, "D_x D_x K"],
    adapted_observation: Float[Array, "D_z D_z K"],
    noise: NumPyNoiseCovariances,
) -> NumPyNoiseCovariances:
    use_adapted = buffer_full & valid
    valid_adapted_count = max(np.sum(use_adapted), 1)

    mean_process = (
        np.sum(
            np.where(
                use_adapted[np.newaxis, np.newaxis, :],
                adapted_process,
                0.0,
            ),
            axis=2,
        )
        / valid_adapted_count
    )

    mean_observation = (
        np.sum(
            np.where(
                use_adapted[np.newaxis, np.newaxis, :],
                adapted_observation,
                0.0,
            ),
            axis=2,
        )
        / valid_adapted_count
    )

    has_valid_adapted = np.any(use_adapted)

    return NumPyNoiseCovariances(
        process_noise_covariance=np.where(
            has_valid_adapted,
            mean_process,
            noise.process_noise_covariance,
        ),
        observation_noise_covariance=np.where(
            has_valid_adapted,
            mean_observation,
            noise.observation_noise_covariance,
        ),
    )


def compute_kalman_gain(
    *,
    predicted_covariance: Float[Array, "... D_x D_x"],
    observation_matrix: Float[Array, "D_z D_x"],
    observation_noise_covariance: Float[Array, "D_z D_z"],
) -> Float[Array, "... D_x D_z"]:
    S = (
        observation_matrix @ predicted_covariance @ observation_matrix.T
        + observation_noise_covariance
    )
    return np.linalg.solve(S, observation_matrix @ predicted_covariance).swapaxes(
        -2, -1
    )


def enforce_spd(matrix: Float[Array, "... N N"]) -> Float[Array, "... N N"]:
    eps = 1e-8
    symmetrised = (matrix + matrix.swapaxes(-2, -1)) / 2
    eigenvalues, eigenvectors = np.linalg.eigh(symmetrised)
    return (
        eigenvectors * np.maximum(eigenvalues, eps)[..., np.newaxis, :]
    ) @ eigenvectors.swapaxes(-2, -1)


def clamp_eigenvalues(
    matrix: Float[Array, "N N"], *, floor: float, ceiling: float
) -> Float[Array, "N N"]:
    symmetrised = (matrix + matrix.T) / 2
    eigenvalues, eigenvectors = np.linalg.eigh(symmetrised)
    clamped = np.clip(eigenvalues, floor, ceiling)
    return (eigenvectors * clamped[np.newaxis, :]) @ eigenvectors.T



from .simple import (
    NumPySimpleState as NumPySimpleState,
    NumPySimpleStateSequence as NumPySimpleStateSequence,
    NumPySimpleStateBatch as NumPySimpleStateBatch,
    NumPySimpleControlInputSequence as NumPySimpleControlInputSequence,
    NumPySimpleControlInputBatch as NumPySimpleControlInputBatch,
    NumPySimpleCosts as NumPySimpleCosts,
    NumPySimpleSampledObstacleStates as NumPySimpleSampledObstacleStates,
    NumPySimpleObstacleStatesForTimeStep as NumPySimpleObstacleStatesForTimeStep,
    NumPySimpleObstacleStates as NumPySimpleObstacleStates,
    JaxSimpleState as JaxSimpleState,
    JaxSimpleStateSequence as JaxSimpleStateSequence,
    JaxSimpleStateBatch as JaxSimpleStateBatch,
    JaxSimpleControlInputSequence as JaxSimpleControlInputSequence,
    JaxSimpleControlInputBatch as JaxSimpleControlInputBatch,
    JaxSimpleCosts as JaxSimpleCosts,
    JaxSimpleSampledObstacleStates as JaxSimpleSampledObstacleStates,
    JaxSimpleObstacleStatesForTimeStep as JaxSimpleObstacleStatesForTimeStep,
    JaxSimpleObstacleStates as JaxSimpleObstacleStates,
)
from .augmented import (
    NumPyAugmentedState as NumPyAugmentedState,
    NumPyAugmentedStateSequence as NumPyAugmentedStateSequence,
    NumPyAugmentedStateBatch as NumPyAugmentedStateBatch,
    NumPyAugmentedControlInputSequence as NumPyAugmentedControlInputSequence,
    NumPyAugmentedControlInputBatch as NumPyAugmentedControlInputBatch,
    JaxAugmentedState as JaxAugmentedState,
    JaxAugmentedStateSequence as JaxAugmentedStateSequence,
    JaxAugmentedStateBatch as JaxAugmentedStateBatch,
    JaxAugmentedControlInputSequence as JaxAugmentedControlInputSequence,
    JaxAugmentedControlInputBatch as JaxAugmentedControlInputBatch,
    AugmentedModel as AugmentedModel,
    AugmentedSampler as AugmentedSampler,
    AugmentedMppi as AugmentedMppi,
    NumPyAugmentedMppi as NumPyAugmentedMppi,
    JaxAugmentedMppi as JaxAugmentedMppi,
    extract as extract,
)



from typing import Protocol, runtime_checkable

from faran.types import (
    jaxtyped,
    JaxGaussianBelief,
    JaxNoiseCovariances,
    JaxNoiseModelProvider,
)
from faran.filters.noise import IdentityNoiseModelProvider
from faran.filters.kf import jax_kalman_filter

from jaxtyping import Array as JaxArray, Float

import equinox as eqx
import jax
import jax.numpy as jnp


@runtime_checkable
class StateTransitionFunction(Protocol):
    def __call__(self, state: Float[JaxArray, "D_x K"]) -> Float[JaxArray, "D_x K"]:
        """Applies the state transition function to the state."""
        ...


class JaxUnscentedKalmanFilter(eqx.Module):
    """Unscented Kalman Filter for nonlinear systems with linear observations."""

    alpha: float = eqx.field(static=True)
    beta: float = eqx.field(static=True)
    noise_model: JaxNoiseModelProvider

    @staticmethod
    def create(
        *,
        alpha: float = 1.0,
        beta: float = 2.0,
        noise_model: JaxNoiseModelProvider | None = None,
    ) -> "JaxUnscentedKalmanFilter":
        """Create an Unscented Kalman Filter.

        Args:
            alpha: Controls the spread of sigma points around the mean.
            beta: Incorporates prior knowledge (2 is used for Gaussian).
            noise_model: An optional noise model provider for adaptive noise estimation.
        """
        return JaxUnscentedKalmanFilter(
            alpha=alpha,
            beta=beta,
            noise_model=IdentityNoiseModelProvider()
            if noise_model is None
            else noise_model,
        )

    @eqx.filter_jit
    @jaxtyped
    def filter(
        self,
        observations: Float[JaxArray, "T D_z K"],
        *,
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
        state_transition: StateTransitionFunction,
        process_noise_covariance: Float[JaxArray, "D_x D_x"],
        observation_noise_covariance: Float[JaxArray, "D_z D_z"],
        observation_matrix: Float[JaxArray, "D_z D_x"],
    ) -> JaxGaussianBelief:
        """Run the UKF over a sequence of observations.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Σ₀ matrix representing initial state uncertainty.
            state_transition: Nonlinear state transition function.
            process_noise_covariance: R matrix representing the covariance of process noise.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            observation_matrix: H matrix mapping state to observation space.
        """
        belief = self.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )
        noise = JaxNoiseCovariances(
            process_noise_covariance, observation_noise_covariance
        )
        adapt = self.noise_model(
            obstacle_count=observations.shape[2],
            observation_matrix=observation_matrix,
            noise=noise,
        )
        noise_state = adapt.state

        def step(carry, observation):
            belief, noise, noise_state = carry
            belief = self.predict(
                belief=belief,
                state_transition=state_transition,
                process_noise_covariance=noise.process_noise_covariance,
            )
            noise, noise_state = adapt(
                noise=noise,
                prediction=belief,
                observation=observation,
                state=noise_state,
            )
            belief = self.update(
                observation,
                prediction=belief,
                observation_matrix=observation_matrix,
                observation_noise_covariance=noise.observation_noise_covariance,
                initial_state_covariance=initial_state_covariance,
            )
            return (belief, noise, noise_state), None

        (belief, _, _), _ = jax.lax.scan(
            step, (belief, noise, noise_state), observations
        )
        return belief

    @eqx.filter_jit
    @jaxtyped
    def predict(
        self,
        *,
        belief: JaxGaussianBelief,
        state_transition: StateTransitionFunction,
        process_noise_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Performs the prediction step of the UKF using the unscented transform.

        Args:
            belief: The current belief about the state.
            state_transition: Nonlinear state transition function.
            process_noise_covariance: R matrix representing the covariance of process noise.
        """
        R = process_noise_covariance
        mu, sigma = belief

        state_dimension = mu.shape[0]
        lambda_ = self.scaling_parameter_for(state_dimension)

        mean_weights, covariance_weights = self._compute_weights(state_dimension)

        def should_skip(
            *,
            mean: Float[JaxArray, "D_x K"],
            covariance: Float[JaxArray, "D_x D_x K"],
        ) -> Float[JaxArray, " K"]:
            return jnp.any(jnp.isnan(covariance), axis=(0, 1)) | jnp.any(
                jnp.isnan(mean), axis=0
            )

        def substitute_missing_values(
            *,
            mean: Float[JaxArray, "D_x K"],
            covariance: Float[JaxArray, "D_x D_x K"],
        ) -> tuple[Float[JaxArray, "D_x K"], Float[JaxArray, "D_x D_x K"]]:
            covariance_is_nan = should_skip(mean=mean, covariance=covariance)

            return (
                jnp.where(jnp.isnan(mean), 0.0, mean),
                jnp.where(
                    covariance_is_nan[jnp.newaxis, jnp.newaxis, :],
                    jnp.eye(state_dimension)[..., jnp.newaxis] * 1e-4,
                    covariance,
                ),
            )

        def predict_single(
            mu_k: JaxArray, sigma_k: JaxArray
        ) -> tuple[JaxArray, JaxArray]:
            sigma_points = self._generate_sigma_points(
                mu_k, sigma_k, state_dimension, lambda_
            )

            def propagate_sigma_point(sigma_point: JaxArray) -> JaxArray:
                return state_transition(sigma_point[:, jnp.newaxis]).flatten()

            propagated_sigma_points = jax.vmap(propagate_sigma_point)(sigma_points)

            predicted_mean_k = jnp.sum(
                mean_weights[:, jnp.newaxis] * propagated_sigma_points, axis=0
            )

            deviations = propagated_sigma_points - predicted_mean_k
            predicted_covariance_k = (
                jnp.einsum("i,ij,ik->jk", covariance_weights, deviations, deviations)
                + R
            )

            return predicted_mean_k, predicted_covariance_k

        def restore_missing(
            *,
            skip_mask: Float[JaxArray, " K"],
            predicted_mean: Float[JaxArray, "D_x K"],
            predicted_covariance: Float[JaxArray, "D_x D_x K"],
        ) -> JaxGaussianBelief:
            return JaxGaussianBelief(
                mean=jnp.where(skip_mask, jnp.nan, predicted_mean),
                covariance=jnp.where(
                    skip_mask[jnp.newaxis, jnp.newaxis, :],
                    jnp.nan,
                    predicted_covariance,
                ),
            )

        skip_mask = should_skip(mean=mu, covariance=sigma)
        safe_mu, safe_sigma = substitute_missing_values(mean=mu, covariance=sigma)

        predicted_means, predicted_covariances = jax.vmap(
            predict_single, in_axes=(1, 2), out_axes=(1, 2)
        )(safe_mu, safe_sigma)

        return restore_missing(
            skip_mask=skip_mask,
            predicted_mean=predicted_means,
            predicted_covariance=predicted_covariances,
        )

    @eqx.filter_jit
    @jaxtyped
    def update(
        self,
        observation: Float[JaxArray, "D_z K"],
        *,
        prediction: JaxGaussianBelief,
        observation_matrix: Float[JaxArray, "D_z D_x"],
        observation_noise_covariance: Float[JaxArray, "D_z D_z"],
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Performs the update step of the UKF using a linear observation model.

        Args:
            observation: The newly observed state.
            prediction: The predicted belief from the prediction step.
            observation_matrix: H matrix mapping state to observation space.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return jax_kalman_filter.update(
            observation=observation,
            prediction=prediction,
            observation_matrix=observation_matrix,
            observation_noise_covariance=observation_noise_covariance,
            initial_state_covariance=initial_state_covariance,
        )

    @eqx.filter_jit
    @jaxtyped
    def initial_belief_from(
        self,
        observations: Float[JaxArray, "T D_z K"],
        *,
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Initializes the belief state from the first observation using a pseudo-inverse.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return jax_kalman_filter.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )

    def scaling_parameter_for(self, state_dimension: int) -> float:
        """Returns the scaling parameter for the given state dimension."""
        return (self.alpha**2 - 1) * state_dimension

    def _compute_weights(self, state_dimension: int) -> tuple[JaxArray, JaxArray]:
        n = state_dimension
        sigma_point_count = 2 * n + 1
        lambda_ = self.scaling_parameter_for(state_dimension)
        alpha = self.alpha
        beta = self.beta

        mean_weights = jnp.zeros(sigma_point_count)
        covariance_weights = jnp.zeros(sigma_point_count)

        mean_weights = mean_weights.at[0].set(lambda_ / (n + lambda_))
        covariance_weights = covariance_weights.at[0].set(
            lambda_ / (n + lambda_) + (1 - alpha**2 + beta)
        )

        remaining_weight = 1 / (2 * (n + lambda_))
        mean_weights = mean_weights.at[1:].set(remaining_weight)
        covariance_weights = covariance_weights.at[1:].set(remaining_weight)

        return mean_weights, covariance_weights

    @staticmethod
    def _generate_sigma_points(
        mean: JaxArray, covariance: JaxArray, state_dimension: int, lambda_: float
    ) -> JaxArray:
        n = state_dimension
        sigma_point_count = 2 * n + 1

        scaled_covariance = (n + lambda_) * covariance

        # NOTE: This is to ensure the scaled covariance is positive definite.
        scaled_covariance = (scaled_covariance + scaled_covariance.T) / 2
        eigenvalues, eigenvectors = jnp.linalg.eigh(scaled_covariance)

        # NOTE: The eigenvalue floor must be large enough to survive the float32
        # round-trip through V @ D @ V^T without losing positive-definiteness.
        eigenvalue_floor = jnp.maximum(1e-10, jnp.max(eigenvalues) * 1e-6)
        eigenvalues = jnp.maximum(eigenvalues, eigenvalue_floor)
        scaled_covariance = eigenvectors @ jnp.diag(eigenvalues) @ eigenvectors.T

        sqrt_scaled_covariance = jnp.linalg.cholesky(scaled_covariance)

        sigma_points = jnp.zeros((sigma_point_count, n))
        sigma_points = sigma_points.at[0].set(mean)

        for i in range(n):
            sigma_points = sigma_points.at[i + 1].set(
                mean + sqrt_scaled_covariance[:, i]
            )
            sigma_points = sigma_points.at[n + i + 1].set(
                mean - sqrt_scaled_covariance[:, i]
            )

        return sigma_points



from typing import Protocol, TypedDict, Mapping

from faran.types.array import Array, DataType

from jaxtyping import Float


type BoundaryPoints = Float[Array, "L 2"]
type Breakpoint = float
type BoundaryWidthsDescription = Mapping[Breakpoint, "BoundaryWidths"]
"""A mapping describing the widths of the boundary at various segments
along a reference trajectory.

Example:
    If the corridor width changes at longitudinal positions 0.0, 10.0, and 20.0
    along the reference trajectory, the mapping could be defined as:
    {
        0.0: {"left": 2.0, "right": 2.0},
        10.0: {"left": 1.5, "right": 2.5},
        20.0: {"left": 2.0, "right": 1.0},
    }
"""


class BoundaryWidths(TypedDict):
    left: float
    """Distance to the left boundary at a breakpoint."""

    right: float
    """Distance to the right boundary at a breakpoint."""


class BoundaryDistance(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        """Returns the distances between the ego and the nearest boundary as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon over which the distances are defined."""
        ...

    @property
    def rollout_count(self) -> int:
        """The number of rollouts for which the distances are defined."""
        ...


class BoundaryDistanceExtractor[StateBatchT, DistanceT](Protocol):
    def __call__(self, *, states: StateBatchT) -> DistanceT:
        """Computes the distances between the ego and the nearest boundary."""
        ...


class ExplicitBoundary(Protocol):
    def left(self, *, sample_count: int = 100) -> BoundaryPoints:
        """Returns the left boundary points as a NumPy array."""
        ...

    def right(self, *, sample_count: int = 100) -> BoundaryPoints:
        """Returns the right boundary points as a NumPy array."""
        ...



from typing import cast
from dataclasses import dataclass

from faran.types import NumPyObstacleStateSampler
from faran.obstacles.basic import NumPySampledObstacle2dPoses, NumPyObstacle2dPoses

from riskit import distribution

import numpy as np


type Rng = np.random.Generator


@dataclass(frozen=True)
class NumPyGaussianObstacle2dPoseSampler(
    NumPyObstacleStateSampler[NumPyObstacle2dPoses, NumPySampledObstacle2dPoses]
):
    """Samples obstacle poses from a Gaussian distribution parameterized by predicted covariances."""

    rng: Rng

    @staticmethod
    def create(*, seed: int = 42) -> "NumPyGaussianObstacle2dPoseSampler":
        return NumPyGaussianObstacle2dPoseSampler(rng=np.random.default_rng(seed))

    def __call__(
        self, states: NumPyObstacle2dPoses, *, count: int
    ) -> NumPySampledObstacle2dPoses:
        if states.count == 0:
            return cast(NumPySampledObstacle2dPoses, states.single())

        if (covariance := states.covariance()) is None:
            assert count == 1, (
                "It's pointless to take multiple samples, when covariance information is not available."
            )
            return cast(NumPySampledObstacle2dPoses, states.single())

        T, D_O, _, K = covariance.shape

        mean = np.stack([states.x(), states.y(), states.heading()], axis=1)
        flat_covariance = covariance.transpose(0, 3, 1, 2).reshape(-1, D_O, D_O)
        flat_mean = mean.transpose(0, 2, 1).reshape(-1, D_O)

        samples = distribution.numpy.gaussian(
            mean=flat_mean, covariance=flat_covariance, rng=self.rng
        ).sample(count=count)

        samples = samples.reshape(T, K, D_O, count).transpose(0, 2, 1, 3)

        return states.sampled(
            x=samples[:, 0, :, :], y=samples[:, 1, :, :], heading=samples[:, 2, :, :]
        )



from .basic import (
    NumPyObstacleIds as NumPyObstacleIds,
    NumPyObstacleStatesRunningHistory as NumPyObstacleStatesRunningHistory,
)
from .accelerated import (
    JaxObstacleIds as JaxObstacleIds,
    JaxObstacleStatesRunningHistory as JaxObstacleStatesRunningHistory,
)



from dataclasses import dataclass, field

from faran.types import ObstacleStateObserver, ObstacleStateProvider
from faran.collectors.access import access
from faran.collectors.registry import (
    DataTransformer,
    IdentityTransformer,
    ModificationNotifierMixin,
    ListCollectorMixin,
    OnModifyCallback,
)


@dataclass(frozen=True)
class ObstacleStateCollector[ObstacleStatesForTimeStepT](
    ObstacleStateObserver[ObstacleStatesForTimeStepT],
    ModificationNotifierMixin,
    ListCollectorMixin,
):
    """Decorator collecting obstacle states at each time step."""

    inner: ObstacleStateObserver[ObstacleStatesForTimeStepT]
    transformer: DataTransformer[ObstacleStatesForTimeStepT]
    _callbacks: list[OnModifyCallback] = field(default_factory=list)
    _collected: list[ObstacleStatesForTimeStepT] = field(default_factory=list)

    @staticmethod
    def decorating[OS](
        observer: ObstacleStateObserver[OS],
        *,
        transformer: DataTransformer[OS] = IdentityTransformer(),
    ) -> "ObstacleStateCollector[OS]":
        return ObstacleStateCollector(observer, transformer=transformer)

    def observe(self, states: ObstacleStatesForTimeStepT) -> None:
        self._collected.append(states)
        self.inner.observe(states)
        self.notify()

    @property
    def key(self) -> str:
        return access.obstacle_states.key


@dataclass(frozen=True)
class ObstacleObservationCollector[ObstacleStatesForTimeStepT](
    ObstacleStateObserver[ObstacleStatesForTimeStepT],
    ModificationNotifierMixin,
    ListCollectorMixin,
):
    """Decorator collecting observed obstacle states at each time step. You can use
    this to collect post-processed observations (e.g., after adding synthetic noise)."""

    inner: ObstacleStateObserver[ObstacleStatesForTimeStepT]
    transformer: DataTransformer[ObstacleStatesForTimeStepT]
    _callbacks: list[OnModifyCallback] = field(default_factory=list)
    _collected: list[ObstacleStatesForTimeStepT] = field(default_factory=list)

    @staticmethod
    def decorating[OS](
        observer: ObstacleStateObserver[OS],
        *,
        transformer: DataTransformer[OS] = IdentityTransformer(),
    ) -> "ObstacleObservationCollector[OS]":
        return ObstacleObservationCollector(observer, transformer=transformer)

    def observe(self, states: ObstacleStatesForTimeStepT) -> None:
        self._collected.append(states)
        self.inner.observe(states)
        self.notify()

    @property
    def key(self) -> str:
        return access.obstacle_observations.key


@dataclass(frozen=True)
class ObstacleForecastCollector[ObstacleStatesT](
    ObstacleStateProvider[ObstacleStatesT],
    ModificationNotifierMixin,
    ListCollectorMixin,
):
    """Decorator collecting predicted obstacle state forecasts."""

    inner: ObstacleStateProvider[ObstacleStatesT]
    transformer: DataTransformer[ObstacleStatesT]
    _callbacks: list[OnModifyCallback] = field(default_factory=list)
    _collected: list[ObstacleStatesT] = field(default_factory=list)

    @staticmethod
    def decorating[OS](
        provider: ObstacleStateProvider[OS],
        *,
        transformer: DataTransformer[OS] = IdentityTransformer(),
    ) -> "ObstacleForecastCollector[OS]":
        return ObstacleForecastCollector(provider, transformer=transformer)

    def __call__(self) -> ObstacleStatesT:
        states = self.inner()
        self._collected.append(states)
        self.notify()
        return states

    @property
    def key(self) -> str:
        return access.obstacle_forecasts.key



from typing import Any
from dataclasses import dataclass

from faran.types import (
    jaxtyped,
    Array,
    StateSequence,
    SimulationData,
    Metric,
    Positions,
    PositionExtractor,
    Trajectory,
    LateralPositions,
)
from faran.collectors import access

from jaxtyping import Float

import numpy as np


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class ComfortMetricResult:
    """Results of the comfort metric, including lateral acceleration and jerk.

    Attributes:
        lateral_acceleration: Lateral acceleration at each time step.
        lateral_jerk: Lateral jerk at each time step.
    """

    lateral_acceleration: Float[Array, " T"]
    lateral_jerk: Float[Array, " T"]


@dataclass(kw_only=True, frozen=True)
class ComfortMetric[StateBatchT, PositionsT, LateralT](Metric[ComfortMetricResult]):
    """Metric evaluating lateral acceleration and jerk relative to a reference trajectory."""

    reference: Trajectory[Any, Any, PositionsT, LateralT]
    time_step_size: float
    position_extractor: PositionExtractor[StateBatchT, PositionsT]

    @staticmethod
    def create[S, P: Positions, L: LateralPositions](
        *,
        reference: Trajectory[Any, Any, P, L],
        time_step_size: float,
        position_extractor: PositionExtractor[S, P],
    ) -> "ComfortMetric[S, P, L]":
        assert time_step_size > 0, (
            f"Time step size must be positive, got {time_step_size}"
        )

        return ComfortMetric(
            reference=reference,
            time_step_size=time_step_size,
            position_extractor=position_extractor,
        )

    def compute(self, data: SimulationData) -> ComfortMetricResult:
        states = data(access.states.assume(StateSequence[StateBatchT]).require())
        state_batch = states.batched()

        positions = self.position_extractor(state_batch)
        lateral = self.reference.lateral(positions)

        lateral_array = np.asarray(lateral)[:, 0]
        lateral_velocity = checked_gradient(lateral_array, self.time_step_size)
        lateral_acceleration = checked_gradient(lateral_velocity, self.time_step_size)
        lateral_jerk = checked_gradient(lateral_acceleration, self.time_step_size)

        return ComfortMetricResult(
            lateral_acceleration=lateral_acceleration, lateral_jerk=lateral_jerk
        )

    @property
    def name(self) -> str:
        return "comfort"


def checked_gradient(y: Float[Array, " L"], dx: float) -> Float[Array, " L"]:
    if y.shape[0] < 2:
        return np.zeros_like(y)

    return np.gradient(y, dx)



from typing import Protocol, runtime_checkable

from faran.types import (
    jaxtyped,
    JaxGaussianBelief,
    JaxNoiseCovariances,
    JaxNoiseModelProvider,
)
from faran.filters.noise import IdentityNoiseModelProvider
from faran.filters.kf import jax_kalman_filter

from jaxtyping import Array as JaxArray, Float

import equinox as eqx
import jax
import jax.numpy as jnp


@runtime_checkable
class StateTransitionFunction(Protocol):
    def __call__(self, state: Float[JaxArray, "D_x K"]) -> Float[JaxArray, "D_x K"]:
        """Applies the state transition function to the state."""
        ...

    def jacobian(self, state: Float[JaxArray, "D_x K"]) -> Float[JaxArray, "D_x D_x K"]:
        """Returns the Jacobian of the state transition function at the given state."""
        ...


class JaxExtendedKalmanFilter(eqx.Module):
    """Extended Kalman Filter for nonlinear systems with linear observations."""

    noise_model: JaxNoiseModelProvider

    @staticmethod
    def create(
        *, noise_model: JaxNoiseModelProvider | None = None
    ) -> "JaxExtendedKalmanFilter":
        return JaxExtendedKalmanFilter(
            noise_model=IdentityNoiseModelProvider()
            if noise_model is None
            else noise_model
        )

    @eqx.filter_jit
    @jaxtyped
    def filter(
        self,
        observations: Float[JaxArray, "T D_z K"],
        *,
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
        state_transition: StateTransitionFunction,
        process_noise_covariance: Float[JaxArray, "D_x D_x"],
        observation_noise_covariance: Float[JaxArray, "D_z D_z"],
        observation_matrix: Float[JaxArray, "D_z D_x"],
    ) -> JaxGaussianBelief:
        """Run the EKF over a sequence of observations.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
            state_transition: Nonlinear differentiable state transition function.
            process_noise_covariance: R matrix representing the covariance of process noise.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            observation_matrix: H matrix mapping state to observation space.
        """
        belief = self.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )
        noise = JaxNoiseCovariances(
            process_noise_covariance, observation_noise_covariance
        )
        adapt = self.noise_model(
            obstacle_count=observations.shape[2],
            observation_matrix=observation_matrix,
            noise=noise,
        )
        noise_state = adapt.state

        def step(carry, observation):
            belief, noise, noise_state = carry
            belief = self.predict(
                belief=belief,
                state_transition=state_transition,
                process_noise_covariance=noise.process_noise_covariance,
            )
            noise, noise_state = adapt(
                noise=noise,
                prediction=belief,
                observation=observation,
                state=noise_state,
            )
            belief = self.update(
                observation,
                prediction=belief,
                observation_matrix=observation_matrix,
                observation_noise_covariance=noise.observation_noise_covariance,
                initial_state_covariance=initial_state_covariance,
            )
            return (belief, noise, noise_state), None

        (belief, _, _), _ = jax.lax.scan(
            step, (belief, noise, noise_state), observations
        )
        return belief

    @eqx.filter_jit
    @jaxtyped
    def predict(
        self,
        *,
        belief: JaxGaussianBelief,
        state_transition: StateTransitionFunction,
        process_noise_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Performs the prediction step of the EKF from the existing belief.

        Args:
            belief: The current belief about the state.
            state_transition: Nonlinear differentiable state transition function.
            process_noise_covariance: R matrix representing the covariance of process noise.
        """
        R = process_noise_covariance
        mu, sigma = belief

        F = state_transition.jacobian(mu)

        return JaxGaussianBelief(
            mean=state_transition(mu),
            covariance=jnp.einsum("ijk,jlk,mlk->imk", F, sigma, F)
            + R[..., jnp.newaxis],
        )

    @eqx.filter_jit
    @jaxtyped
    def update(
        self,
        observation: Float[JaxArray, "D_z K"],
        *,
        prediction: JaxGaussianBelief,
        observation_matrix: Float[JaxArray, "D_z D_x"],
        observation_noise_covariance: Float[JaxArray, "D_z D_z"],
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Performs the update step of the EKF using a new observation.

        Args:
            observation: The newly observed state.
            prediction: The predicted belief from the prediction step.
            observation_matrix: H matrix mapping state to observation space.
            observation_noise_covariance: Q matrix representing the covariance of observation noise.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return jax_kalman_filter.update(
            observation=observation,
            prediction=prediction,
            observation_matrix=observation_matrix,
            observation_noise_covariance=observation_noise_covariance,
            initial_state_covariance=initial_state_covariance,
        )

    @eqx.filter_jit
    @jaxtyped
    def initial_belief_from(
        self,
        observations: Float[JaxArray, "T D_z K"],
        *,
        initial_state_covariance: Float[JaxArray, "D_x D_x"],
    ) -> JaxGaussianBelief:
        """Initializes the belief state from the first observation using a pseudo-inverse.

        Args:
            observations: The observed state history up to the current time step.
            initial_state_covariance: Sigma_0 matrix representing initial state uncertainty.
        """
        return jax_kalman_filter.initial_belief_from(
            observations, initial_state_covariance=initial_state_covariance
        )



from .array import (
    Array as Array,
    DataType as DataType,
    Device as Device,
    HasShape as HasShape,
    jaxtyped as jaxtyped,
    place as place,
)
from .mppi import (
    State as State,
    StateSequence as StateSequence,
    StateBatch as StateBatch,
    ControlInputSequence as ControlInputSequence,
    ControlInputBatch as ControlInputBatch,
    DynamicalModel as DynamicalModel,
    Costs as Costs,
    CostFunction as CostFunction,
    Sampler as Sampler,
    UpdateFunction as UpdateFunction,
    PaddingFunction as PaddingFunction,
    FilterFunction as FilterFunction,
    Weights as Weights,
    Control as Control,
    DebugData as DebugData,
    Mppi as Mppi,
    NumPyState as NumPyState,
    NumPyStateSequence as NumPyStateSequence,
    NumPyStateBatch as NumPyStateBatch,
    NumPyControlInputSequence as NumPyControlInputSequence,
    NumPyControlInputBatch as NumPyControlInputBatch,
    NumPyDynamicalModel as NumPyDynamicalModel,
    NumPyCosts as NumPyCosts,
    NumPySampler as NumPySampler,
    NumPyCostFunction as NumPyCostFunction,
    NumPyUpdateFunction as NumPyUpdateFunction,
    NumPyPaddingFunction as NumPyPaddingFunction,
    NumPyFilterFunction as NumPyFilterFunction,
    JaxState as JaxState,
    JaxStateSequence as JaxStateSequence,
    JaxStateBatch as JaxStateBatch,
    JaxControlInputSequence as JaxControlInputSequence,
    JaxControlInputBatch as JaxControlInputBatch,
    JaxDynamicalModel as JaxDynamicalModel,
    JaxCosts as JaxCosts,
    JaxCostFunction as JaxCostFunction,
    JaxSampler as JaxSampler,
    JaxUpdateFunction as JaxUpdateFunction,
    JaxPaddingFunction as JaxPaddingFunction,
    JaxFilterFunction as JaxFilterFunction,
)
from .models import (
    BICYCLE_D_X as BICYCLE_D_X,
    BICYCLE_D_U as BICYCLE_D_U,
    BICYCLE_D_O as BICYCLE_D_O,
    BicycleD_x as BicycleD_x,
    BicycleD_u as BicycleD_u,
    BicycleD_o as BicycleD_o,
    BicycleState as BicycleState,
    BicycleStateSequence as BicycleStateSequence,
    BicycleStateBatch as BicycleStateBatch,
    BicyclePositions as BicyclePositions,
    BicycleControlInputSequence as BicycleControlInputSequence,
    BicycleControlInputBatch as BicycleControlInputBatch,
    NumPyBicycleObstacleStatesHistory as NumPyBicycleObstacleStatesHistory,
    JaxBicycleObstacleStatesHistory as JaxBicycleObstacleStatesHistory,
    UNICYCLE_D_X as UNICYCLE_D_X,
    UNICYCLE_D_U as UNICYCLE_D_U,
    UNICYCLE_D_O as UNICYCLE_D_O,
    UnicycleD_x as UnicycleD_x,
    UnicycleD_u as UnicycleD_u,
    UnicycleD_o as UnicycleD_o,
    UnicycleState as UnicycleState,
    UnicycleStateSequence as UnicycleStateSequence,
    UnicycleStateBatch as UnicycleStateBatch,
    UnicyclePositions as UnicyclePositions,
    UnicycleControlInputSequence as UnicycleControlInputSequence,
    UnicycleControlInputBatch as UnicycleControlInputBatch,
    NumPyUnicycleObstacleStatesHistory as NumPyUnicycleObstacleStatesHistory,
    JaxUnicycleObstacleStatesHistory as JaxUnicycleObstacleStatesHistory,
    IntegratorState as IntegratorState,
    IntegratorStateSequence as IntegratorStateSequence,
    IntegratorStateBatch as IntegratorStateBatch,
    IntegratorControlInputSequence as IntegratorControlInputSequence,
    IntegratorControlInputBatch as IntegratorControlInputBatch,
    NumPyIntegratorState as NumPyIntegratorState,
    NumPyIntegratorStateSequence as NumPyIntegratorStateSequence,
    NumPyIntegratorStateBatch as NumPyIntegratorStateBatch,
    NumPyIntegratorControlInputSequence as NumPyIntegratorControlInputSequence,
    NumPyIntegratorControlInputBatch as NumPyIntegratorControlInputBatch,
    NumPyIntegratorObstacleStatesHistory as NumPyIntegratorObstacleStatesHistory,
    JaxIntegratorState as JaxIntegratorState,
    JaxIntegratorStateSequence as JaxIntegratorStateSequence,
    JaxIntegratorStateBatch as JaxIntegratorStateBatch,
    JaxIntegratorControlInputSequence as JaxIntegratorControlInputSequence,
    JaxIntegratorControlInputBatch as JaxIntegratorControlInputBatch,
    JaxIntegratorObstacleStatesHistory as JaxIntegratorObstacleStatesHistory,
)
from .trajectories import (
    D_r as D_r,
    D_R as D_R,
    Trajectory as Trajectory,
    PathParameters as PathParameters,
    ReferencePoints as ReferencePoints,
    Positions as Positions,
    LateralPositions as LateralPositions,
    LongitudinalPositions as LongitudinalPositions,
    Normals as Normals,
    NumPyPositions as NumPyPositions,
    NumPySimplePositions as NumPySimplePositions,
    NumPyHeadings as NumPyHeadings,
    NumPyPathParameters as NumPyPathParameters,
    NumPyReferencePoints as NumPyReferencePoints,
    NumPyLateralPositions as NumPyLateralPositions,
    NumPyLongitudinalPositions as NumPyLongitudinalPositions,
    NumPyNormals as NumPyNormals,
    JaxPositions as JaxPositions,
    JaxSimplePositions as JaxSimplePositions,
    JaxHeadings as JaxHeadings,
    JaxPathParameters as JaxPathParameters,
    JaxReferencePoints as JaxReferencePoints,
    JaxLateralPositions as JaxLateralPositions,
    JaxLongitudinalPositions as JaxLongitudinalPositions,
    JaxNormals as JaxNormals,
)
from .samplers import (
    NumPyControlInputBatchCreator as NumPyControlInputBatchCreator,
    JaxControlInputBatchCreator as JaxControlInputBatchCreator,
)
from .states import (
    AugmentedState as AugmentedState,
    AugmentedStateSequence as AugmentedStateSequence,
    AugmentedStateBatch as AugmentedStateBatch,
    AugmentedControlInputSequence as AugmentedControlInputSequence,
    AugmentedControlInputBatch as AugmentedControlInputBatch,
    AugmentedStateCreator as AugmentedStateCreator,
    AugmentedStateSequenceCreator as AugmentedStateSequenceCreator,
    AugmentedStateBatchCreator as AugmentedStateBatchCreator,
    AugmentedControlInputBatchCreator as AugmentedControlInputBatchCreator,
    HasPhysical as HasPhysical,
    HasVirtual as HasVirtual,
)
from .costs import (
    Error as Error,
    PositionExtractor as PositionExtractor,
    ContouringCost as ContouringCost,
    LagCost as LagCost,
    CostSumFunction as CostSumFunction,
    NumPyPathParameterExtractor as NumPyPathParameterExtractor,
    NumPyPathVelocityExtractor as NumPyPathVelocityExtractor,
    NumPyPositionExtractor as NumPyPositionExtractor,
    NumPyHeadingExtractor as NumPyHeadingExtractor,
    JaxPathParameterExtractor as JaxPathParameterExtractor,
    JaxPathVelocityExtractor as JaxPathVelocityExtractor,
    JaxPositionExtractor as JaxPositionExtractor,
    JaxHeadingExtractor as JaxHeadingExtractor,
    POSE_D_O as POSE_D_O,
    PoseD_o as PoseD_o,
    SampledObstacleStates as SampledObstacleStates,
    ObstacleStatesForTimeStep as ObstacleStatesForTimeStep,
    ObstacleStates as ObstacleStates,
    SampledObstaclePositions as SampledObstaclePositions,
    SampledObstacleHeadings as SampledObstacleHeadings,
    SampledObstaclePositionExtractor as SampledObstaclePositionExtractor,
    SampledObstacleHeadingExtractor as SampledObstacleHeadingExtractor,
    NumPySampledObstacleStates as NumPySampledObstacleStates,
    NumPySampledObstaclePositions as NumPySampledObstaclePositions,
    NumPySampledObstacleHeadings as NumPySampledObstacleHeadings,
    NumPyObstacleStatesForTimeStep as NumPyObstacleStatesForTimeStep,
    NumPyObstacleStates as NumPyObstacleStates,
    NumPySampledObstaclePositionExtractor as NumPySampledObstaclePositionExtractor,
    NumPySampledObstacleHeadingExtractor as NumPySampledObstacleHeadingExtractor,
    JaxSampledObstacleStates as JaxSampledObstacleStates,
    JaxSampledObstaclePositions as JaxSampledObstaclePositions,
    JaxSampledObstacleHeadings as JaxSampledObstacleHeadings,
    JaxObstacleStatesForTimeStep as JaxObstacleStatesForTimeStep,
    JaxObstacleStates as JaxObstacleStates,
    JaxSampledObstaclePositionExtractor as JaxSampledObstaclePositionExtractor,
    JaxSampledObstacleHeadingExtractor as JaxSampledObstacleHeadingExtractor,
    ObstacleStateProvider as ObstacleStateProvider,
    ObstacleStateSampler as ObstacleStateSampler,
    Distance as Distance,
    DistanceExtractor as DistanceExtractor,
    SampleCostFunction as SampleCostFunction,
    Risk as Risk,
    RiskMetric as RiskMetric,
    NumPyObstacleStateProvider as NumPyObstacleStateProvider,
    NumPyObstacleStateSampler as NumPyObstacleStateSampler,
    NumPyDistanceExtractor as NumPyDistanceExtractor,
    NumPyRisk as NumPyRisk,
    NumPyRiskMetric as NumPyRiskMetric,
    JaxObstacleStateProvider as JaxObstacleStateProvider,
    JaxObstacleStateSampler as JaxObstacleStateSampler,
    JaxDistanceExtractor as JaxDistanceExtractor,
    JaxRisk as JaxRisk,
    JaxRiskMetric as JaxRiskMetric,
    BoundaryPoints as BoundaryPoints,
    BoundaryDistance as BoundaryDistance,
    BoundaryDistanceExtractor as BoundaryDistanceExtractor,
    ExplicitBoundary as ExplicitBoundary,
    BoundaryWidthsDescription as BoundaryWidthsDescription,
    BoundaryWidths as BoundaryWidths,
    NumPyBoundaryDistance as NumPyBoundaryDistance,
    NumPyBoundaryDistanceExtractor as NumPyBoundaryDistanceExtractor,
    JaxBoundaryDistance as JaxBoundaryDistance,
    JaxBoundaryDistanceExtractor as JaxBoundaryDistanceExtractor,
)
from .predictors import (
    EstimatedObstacleStates as EstimatedObstacleStates,
    ObstacleStatesHistory as ObstacleStatesHistory,
    ObstacleIds as ObstacleIds,
    ObstacleIdAssignment as ObstacleIdAssignment,
    ObstacleStatesRunningHistory as ObstacleStatesRunningHistory,
    ObstacleStateEstimator as ObstacleStateEstimator,
    InputAssumptionProvider as InputAssumptionProvider,
    ObstacleModel as ObstacleModel,
    PredictionCreator as PredictionCreator,
    ObstacleMotionPredictor as ObstacleMotionPredictor,
    NumPyObstacleStatesHistory as NumPyObstacleStatesHistory,
    JaxObstacleStatesHistory as JaxObstacleStatesHistory,
)
from .obstacles import (
    ObstaclePositionsForTimeStep as ObstaclePositionsForTimeStep,
    ObstaclePositions as ObstaclePositions,
    ObstaclePositionExtractor as ObstaclePositionExtractor,
    ObstacleOrientationsForTimeStep as ObstacleOrientationsForTimeStep,
    ObstacleOrientations as ObstacleOrientations,
    ObstacleOrientationExtractor as ObstacleOrientationExtractor,
    ObstacleSimulator as ObstacleSimulator,
    NumPyObstaclePositionsForTimeStep as NumPyObstaclePositionsForTimeStep,
    NumPyObstaclePositions as NumPyObstaclePositions,
    NumPyObstaclePositionExtractor as NumPyObstaclePositionExtractor,
    NumPyObstacleOrientationsForTimeStep as NumPyObstacleOrientationsForTimeStep,
    NumPyObstacleOrientations as NumPyObstacleOrientations,
    NumPyObstacleOrientationExtractor as NumPyObstacleOrientationExtractor,
    NumPyObstacleSimulator as NumPyObstacleSimulator,
    JaxObstaclePositionsForTimeStep as JaxObstaclePositionsForTimeStep,
    JaxObstaclePositions as JaxObstaclePositions,
    JaxObstaclePositionExtractor as JaxObstaclePositionExtractor,
    JaxObstacleOrientationsForTimeStep as JaxObstacleOrientationsForTimeStep,
    JaxObstacleOrientations as JaxObstacleOrientations,
    JaxObstacleOrientationExtractor as JaxObstacleOrientationExtractor,
    JaxObstacleSimulator as JaxObstacleSimulator,
)
from .collectors import (
    SimulationDataAccessor as SimulationDataAccessor,
    SimulationData as SimulationData,
    StateSequenceCreator as StateSequenceCreator,
    ObstacleStateSequencesCreator as ObstacleStateSequencesCreator,
    ObstacleStateObserver as ObstacleStateObserver,
    StateTrajectories as StateTrajectories,
)
from .metrics import (
    Metric as Metric,
)
from .filters import (
    Noise as Noise,
    NoiseModel as NoiseModel,
    NoiseModelProvider as NoiseModelProvider,
    NumPyNoiseCovarianceArrayDescription as NumPyNoiseCovarianceArrayDescription,
    NumPyNoiseCovarianceDescription as NumPyNoiseCovarianceDescription,
    NumPyGaussianBelief as NumPyGaussianBelief,
    NumPyNoiseCovariances as NumPyNoiseCovariances,
    NumPyNoiseModel as NumPyNoiseModel,
    NumPyNoiseModelProvider as NumPyNoiseModelProvider,
    JaxNoiseCovarianceArrayDescription as JaxNoiseCovarianceArrayDescription,
    JaxNoiseCovarianceDescription as JaxNoiseCovarianceDescription,
    JaxGaussianBelief as JaxGaussianBelief,
    JaxNoiseCovariances as JaxNoiseCovariances,
    JaxNoiseModel as JaxNoiseModel,
    JaxNoiseModelProvider as JaxNoiseModelProvider,
)



from typing import cast, Any
from dataclasses import dataclass

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    JaxState,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxDynamicalModel,
    JaxCostFunction,
    JaxSampler,
    JaxUpdateFunction,
    JaxPaddingFunction,
    JaxFilterFunction,
    Mppi,
    DebugData,
    Control,
)
from faran.mppi.common import UseOptimalControlUpdate, NoFilter

from jaxtyping import Array as JaxArray, Float, Scalar

import jax
import jax.numpy as jnp
import numpy as np


@jaxtyped
@dataclass(frozen=True)
class JaxWeights:
    """Softmax-normalized cost weights used to compute the weighted average over rollouts."""

    _array: Float[JaxArray, " M"]

    def __array__(self, dtype: DataType | None = None) -> Float[Array, " M"]:
        return np.asarray(self._array)

    @property
    def rollout_count(self) -> int:
        return self._array.shape[0]

    @property
    def array(self) -> Float[JaxArray, " M"]:
        return self._array


class JaxZeroPadding(
    JaxPaddingFunction[JaxControlInputSequence, JaxControlInputSequence]
):
    """Fills shifted-out time steps with zero control inputs after horizon advancement."""

    def __call__(
        self, *, nominal_input: JaxControlInputSequence, padding_size: int
    ) -> JaxControlInputSequence:
        padding_array = jnp.zeros((padding_size, nominal_input.array.shape[1]))
        return nominal_input.similar(array=padding_array)


@dataclass(kw_only=True, frozen=True)
class JaxMppi[
    StateT: JaxState,
    StateBatchT: JaxStateBatch,
    ControlInputSequenceT: JaxControlInputSequence,
    ControlInputBatchT: JaxControlInputBatch,
    ControlInputPaddingT: JaxControlInputSequence = ControlInputSequenceT,
](Mppi[StateT, ControlInputSequenceT, JaxWeights]):
    """Sampling-based stochastic optimal controller using the information-theoretic MPPI algorithm."""

    planning_interval: int
    model: JaxDynamicalModel[
        StateT, Any, StateBatchT, ControlInputSequenceT, ControlInputBatchT
    ]
    cost_function: JaxCostFunction
    sampler: JaxSampler[ControlInputSequenceT, ControlInputBatchT]
    update_function: JaxUpdateFunction[ControlInputSequenceT]
    padding_function: JaxPaddingFunction[ControlInputSequenceT, ControlInputPaddingT]
    filter_function: JaxFilterFunction[ControlInputSequenceT]

    @staticmethod
    def create[
        S: JaxState,
        SB: JaxStateBatch,
        CIS: JaxControlInputSequence,
        CIB: JaxControlInputBatch,
        CIP: JaxControlInputSequence = CIS,
    ](
        *,
        planning_interval: int = 1,
        model: JaxDynamicalModel[S, Any, SB, CIS, CIB],
        cost_function: JaxCostFunction,
        sampler: JaxSampler[CIS, CIB],
        update_function: JaxUpdateFunction[CIS] | None = None,
        padding_function: JaxPaddingFunction[CIS, CIP] | None = None,
        filter_function: JaxFilterFunction[CIS] | None = None,
    ) -> "JaxMppi[S, SB, CIS, CIB, CIP]":
        return JaxMppi(
            planning_interval=planning_interval,
            model=model,
            cost_function=cost_function,
            sampler=sampler,
            update_function=update_function
            or cast(JaxUpdateFunction[CIS], UseOptimalControlUpdate()),
            padding_function=padding_function
            or cast(JaxPaddingFunction[CIS, CIP], JaxZeroPadding()),
            filter_function=filter_function or cast(JaxFilterFunction[CIS], NoFilter()),
        )

    def __post_init__(self) -> None:
        assert self.planning_interval > 0, "Planning interval must be positive."

    def step(
        self,
        *,
        temperature: float,
        nominal_input: ControlInputSequenceT,
        initial_state: StateT,
    ) -> Control[ControlInputSequenceT, JaxWeights]:
        """Samples rollouts, evaluates costs, and returns the cost-weighted optimal control."""
        assert temperature > 0.0, "Temperature must be positive."

        samples = self.sampler.sample(around=nominal_input)

        rollouts = self.model.simulate(inputs=samples, initial_state=initial_state)
        costs = self.cost_function(inputs=samples, states=rollouts)

        optimal_control, weights = compute_weighted_control(
            samples.array, costs.array, temperature
        )
        optimal_input = self.filter_function(
            optimal_input=nominal_input.similar(array=optimal_control)
        )

        nominal_input = self.update_function(
            nominal_input=nominal_input, optimal_input=optimal_input
        )

        shifted_control = shift_control_left(
            nominal_input.array,
            padding_array=self.padding_function(
                nominal_input=nominal_input, padding_size=self.planning_interval
            ).array,
            planning_interval=self.planning_interval,
        )

        return Control(
            optimal=optimal_input,
            nominal=nominal_input.similar(array=shifted_control),
            debug=DebugData(trajectory_weights=JaxWeights(weights)),
        )


@jax.jit
@jaxtyped
def compute_weighted_control(
    samples: Float[JaxArray, "T D_u M"],
    costs: Float[JaxArray, "T M"],
    temperature: Scalar,
) -> tuple[Float[JaxArray, "T D_u"], Float[JaxArray, " M"]]:
    total_costs = jnp.sum(costs, axis=0)
    min_cost = jnp.min(total_costs)
    exp_costs = jnp.exp((total_costs - min_cost) / (-temperature))
    normalizing_constant = jnp.sum(exp_costs)
    weights = exp_costs / normalizing_constant

    return jnp.tensordot(samples, weights, axes=([2], [0])), weights


@jax.jit(static_argnames=("planning_interval",))
@jaxtyped
def shift_control_left(
    control: Float[JaxArray, "T D_u"],
    *,
    padding_array: Float[JaxArray, "P D_u"],
    planning_interval: int,
) -> Float[JaxArray, "T D_u"]:
    return jnp.concatenate([control[planning_interval:], padding_array], axis=0)



from typing import Final

from faran.predictors.common import StaticPredictor, CurvilinearPredictor


class predictor:
    """Factory namespace for creating obstacle motion predictors."""

    class numpy:
        curvilinear: Final = CurvilinearPredictor.create
        static: Final = StaticPredictor.create

    class jax:
        curvilinear: Final = CurvilinearPredictor.create
        static: Final = StaticPredictor.create



from typing import cast
from dataclasses import dataclass

from faran.types import (
    ObstacleStatesForTimeStep,
    ObstacleStatesHistory,
    ObstacleModel,
    ObstacleStateEstimator,
    PredictionCreator,
    InputAssumptionProvider,
)


@dataclass(kw_only=True, frozen=True)
class StaticPredictor:
    """Predicts obstacle motion by replicating the last observed state over the horizon."""

    horizon: int

    @staticmethod
    def create(*, horizon: int) -> "StaticPredictor":
        return StaticPredictor(horizon=horizon)

    def predict[PredictionT](
        self, *, history: ObstacleStatesHistory[ObstacleStatesForTimeStep[PredictionT]]
    ) -> PredictionT:
        assert history.horizon > 0, (
            "There should be at least one observation in history."
        )

        return history.last().replicate(horizon=self.horizon)


class NoAssumptions[InputsT]:
    """Identity assumption provider that passes inputs through unchanged."""

    def __call__(self, inputs: InputsT, /) -> InputsT:
        return inputs


@dataclass(kw_only=True, frozen=True)
class CurvilinearPredictor[
    HistoryT: ObstacleStatesHistory,
    StatesT,
    InputsT,
    CovarianceT,
    StateSequencesT,
    PredictionT,
]:
    """Predicts obstacle motion by forward propagating estimated states with a curvilinear model."""

    horizon: int
    estimator: ObstacleStateEstimator[HistoryT, StatesT, InputsT, CovarianceT]
    assumptions: InputAssumptionProvider[InputsT]
    model: ObstacleModel[HistoryT, StatesT, InputsT, CovarianceT, StateSequencesT]
    prediction: PredictionCreator[StateSequencesT, PredictionT]

    @staticmethod
    def create[H: ObstacleStatesHistory, S, I, C, SS, P](
        *,
        horizon: int,
        model: ObstacleModel[H, S, I, C, SS],
        estimator: ObstacleStateEstimator[H, S, I, C],
        prediction: PredictionCreator[SS, P],
        assumptions: InputAssumptionProvider[I] | None = None,
    ) -> "CurvilinearPredictor[H, S, I, C, SS, P]":
        assumptions = (
            assumptions
            if assumptions is not None
            else cast(InputAssumptionProvider[I], NoAssumptions())
        )

        return CurvilinearPredictor(
            horizon=horizon,
            estimator=estimator,
            assumptions=assumptions,
            model=model,
            prediction=prediction,
        )

    def predict(self, *, history: HistoryT) -> PredictionT:
        if history.horizon == 0:
            return self.prediction.empty(horizon=self.horizon)

        estimated = self.estimator.estimate_from(history)
        inputs = self.assumptions(estimated.inputs)

        return self.prediction(
            states=self.model.forward(
                states=estimated.states,
                inputs=inputs,
                covariances=estimated.covariance,
                horizon=self.horizon,
            )
        )



from .circles import (
    Circles as Circles,
    NumPyCircleDistanceExtractor as NumPyCircleDistanceExtractor,
    JaxCircleDistanceExtractor as JaxCircleDistanceExtractor,
)
from .sat import (
    NumPySatDistanceExtractor as NumPySatDistanceExtractor,
    JaxSatDistanceExtractor as JaxSatDistanceExtractor,
    ConvexPolygon as ConvexPolygon,
)



from .basic import (
    NumPyKalmanFilter as NumPyKalmanFilter,
    numpy_kalman_filter as numpy_kalman_filter,
)
from .accelerated import (
    JaxKalmanFilter as JaxKalmanFilter,
    jax_kalman_filter as jax_kalman_filter,
)



from typing import Protocol

from faran.types.array import Array
from faran.types.predictors.common import ObstacleStatesHistory

from jaxtyping import Float


class NumPyObstacleStatesHistory[ObstacleStatesForTimeStepT](
    ObstacleStatesHistory[ObstacleStatesForTimeStepT], Protocol
):
    @property
    def array(self) -> Float[Array, "T D_o K"]:
        """Returns the obstacle state history as a NumPy array."""
        ...
